# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import glob
import os
import re
import shlex
import sys
from collections.abc import Sequence
from pathlib import Path

import pmb.build
import pmb.chroot
import pmb.chroot.apk
import pmb.chroot.initfs
import pmb.config
import pmb.config.pmaports
import pmb.helpers.devices
import pmb.helpers.other
import pmb.helpers.run
import pmb.install
import pmb.install.blockdevice
import pmb.install.recovery
import pmb.install.ui
import pmb.parse.depends
from pmb.core import Chroot, ChrootType, Config
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.helpers.locale import get_xkb_layout
from pmb.helpers.mount import mount_device_rootfs
from pmb.parse.deviceinfo import Deviceinfo
from pmb.types import PartitionLayout, PmbArgs, RunOutputTypeDefault

# Keep track of the packages we already visited in get_recommends() to avoid
# infinite recursion
get_recommends_visited: list[str] = []
get_selected_providers_visited: list[str] = []


def get_subpartitions_size(chroot: Chroot) -> tuple[int, int]:
    """
    Calculate the size of the boot and root subpartition.

    :param suffix: the chroot suffix, e.g. "rootfs_qemu-amd64"
    :returns: (boot, root) the size of the boot and root
              partition as integer in MiB
    """
    config = get_context().config
    boot = int(config.boot_size)

    # Estimate root partition size, then add some free space. The size
    # calculation is not as trivial as one may think, and depending on the
    # file system etc it seems to be just impossible to get it right.
    root = pmb.helpers.other.folder_size(chroot.path) / 1024
    root *= 1.20
    root += 50 + int(config.extra_space)
    return (boot, round(root))


def get_nonfree_packages(device: str) -> list[str]:
    """
    Get any legacy non-free subpackages in the APKBUILD.
    Also see: https://postmarketos.org/edge/2024/02/15/default-nonfree-fw/

    :returns: list of non-free packages to be installed. Example:
              ["device-nokia-n900-nonfree-firmware"]
    """
    # Read subpackages
    device_path = pmb.helpers.devices.find_path(device, "APKBUILD")
    if not device_path:
        raise RuntimeError(f"Device package not found for {device}")

    apkbuild = pmb.parse.apkbuild(device_path)
    subpackages = apkbuild["subpackages"]

    # Check for firmware and userland
    ret = []
    prefix = "device-" + device + "-nonfree-"
    if prefix + "firmware" in subpackages:
        ret += [prefix + "firmware"]
    if prefix + "userland" in subpackages:
        ret += [prefix + "userland"]
    return ret


def get_kernel_package(config: Config) -> list[str]:
    """
    Get the device's kernel subpackage based on the user's choice in
    "pmbootstrap init".

    :param device: code name, e.g. "sony-amami"
    :returns: [] or the package in a list, e.g.
              ["device-sony-amami-kernel-mainline"]
    """
    # Empty list: single kernel devices / "none" selected
    kernels = pmb.parse._apkbuild.kernels(config.device)
    if not kernels or config.kernel == "none":
        return []

    # Sanity check
    if config.kernel not in kernels:
        raise NonBugError(
            "Selected kernel (" + config.kernel + ") is not"
            " valid for device " + config.device + ". Please"
            " run 'pmbootstrap init' to select a valid kernel."
        )

    # Selected kernel subpackage
    return ["device-" + config.device + "-kernel-" + config.kernel]


def copy_files_from_chroot(chroot: Chroot, rsync: bool, verbose: bool) -> None:
    """
    Copy all files from the rootfs chroot to /mnt/install, except
    for the home folder (because /home will contain some empty
    mountpoint folders).

    :param suffix: the chroot suffix, e.g. "rootfs_qemu-amd64"
    """
    # Mount the device rootfs
    logging.info(f"(native) copy {chroot} to /mnt/install/")
    mountpoint = mount_device_rootfs(chroot)
    mountpoint_outside = Chroot.native() / mountpoint

    # Remove empty qemu-user binary stub (where the binary was bind-mounted)
    arch_qemu = pmb.parse.deviceinfo().arch.qemu_user()
    qemu_binary = mountpoint_outside / f"usr/bin/qemu-{arch_qemu}-static"
    if os.path.exists(qemu_binary):
        pmb.helpers.run.root(["rm", qemu_binary])

    # Remove apk progress fifo
    fifo = chroot / "tmp/apk_progress_fifo"
    if os.path.exists(fifo):
        pmb.helpers.run.root(["rm", fifo])

    # Get all folders inside the device rootfs (except for home)
    folders: list[str] = []
    for path in mountpoint_outside.glob("*"):
        if path.name == "home":
            continue
        folders.append(path.name)

    # Update or copy all files
    if rsync:
        pmb.chroot.apk.install(["rsync"], Chroot.native())
        rsync_flags = "-a"
        if verbose:
            rsync_flags += "vP"
        pmb.chroot.root(
            ["rsync", rsync_flags, "--delete", *folders, "/mnt/install/"], working_dir=mountpoint
        )
        pmb.chroot.root(["rm", "-rf", "/mnt/install/home"])
    else:
        pmb.chroot.root(["cp", "-a", *folders, "/mnt/install/"], working_dir=mountpoint)

    # Log how much space and inodes we have used
    pmb.chroot.user(["df", "-h", "/mnt/install"])
    pmb.chroot.user(["df", "-i", "/mnt/install"])


def create_home_from_skel(filesystem: str, user: str) -> None:
    """Create /home/{user} from /etc/skel"""
    rootfs = Chroot.native() / "mnt/install"
    # In btrfs, home subvol & home dir is created in format.py
    if filesystem != "btrfs":
        pmb.helpers.run.root(["mkdir", rootfs / "home"])

    home = rootfs / "home" / user
    if (rootfs / "etc/skel").exists():
        pmb.helpers.run.root(["cp", "-a", (rootfs / "etc/skel"), home])
    else:
        pmb.helpers.run.root(["mkdir", home])
    pmb.helpers.run.root(["chown", "-R", "10000", home])


def configure_apk(install_local_pkgs: bool) -> None:
    """
    Copy over all official keys, and the keys used to compile local packages
    (unless --no-local-pkgs is set). Then copy the corresponding APKINDEX files
    and remove the /mnt/pmbootstrap/packages repository.
    """
    # Official keys
    keys_dir = pmb.config.apk_keys_path

    # Official keys + local keys
    if install_local_pkgs:
        keys_dir = get_context().config.work / "config_apk_keys"

    # Copy over keys
    rootfs = Chroot.native() / "mnt/install"
    for key in keys_dir.glob("*.pub"):
        pmb.helpers.run.root(["cp", key, rootfs / "etc/apk/keys/"])

    # Copy over the corresponding APKINDEX files from cache
    index_files = pmb.helpers.repo.apkindex_files(
        arch=pmb.parse.deviceinfo().arch, user_repository=False
    )
    for f in index_files:
        pmb.helpers.run.root(["cp", f, rootfs / "var/cache/apk/"])

    # Disable pmbootstrap repository
    pmb.chroot.root(
        ["sed", "-i", r"/\/mnt\/pmbootstrap\/packages/d", "/mnt/install/etc/apk/repositories"]
    )
    pmb.helpers.run.user(["cat", rootfs / "etc/apk/repositories"])


def set_user(config: Config) -> None:
    """
    Create user with UID 10000 if it doesn't exist.
    Usually the ID for the first user created is 1000, but higher ID is
    chosen here to not cause issues with existing installations. Historically,
    this was done to avoid conflict with Android UIDs/GIDs, but pmOS has since
    dropped support for hybris/Halium.
    """
    chroot = Chroot.rootfs(config.device)
    if not pmb.chroot.user_exists(config.user, chroot):
        pmb.chroot.root(["adduser", "-D", "-u", "10000", config.user], chroot)

    pmaports_cfg = pmb.config.pmaports.read_config()
    groups = []
    groups += pmaports_cfg.get(
        "install_user_groups", "audio,input,netdev,plugdev,video,wheel"
    ).split(",")
    groups += pmb.install.ui.get_groups(config)

    for group in groups:
        pmb.chroot.root(["addgroup", "-S", group], chroot, check=False)
        pmb.chroot.root(["addgroup", config.user, group], chroot)


def setup_login_chpasswd_user_from_arg(user: str, password: str, chroot: Chroot) -> None:
    """
    Set the user's password from what the user passed as --password. Make an
    effort to not have the password end up in the log file by writing it to
    a temp file, instead of "echo user:$pass | chpasswd". The user should of
    course only use this with a test password anyway, but let's be nice and try
    to have the user protected from accidentally posting their password in
    any case.

    :param suffix: of the chroot, where passwd will be execute (either the
                   rootfs_{args.device} or installer_{args.device}
    """
    path = "/tmp/pmbootstrap_chpasswd_in"
    path_outside = chroot / path

    path_outside.write_text(f"{user}:{password}", encoding="utf-8")

    pmb.chroot.root(["sh", "-c", f"cat {shlex.quote(path)} | chpasswd"], chroot)

    os.unlink(path_outside)


def is_root_locked(chroot: Chroot) -> bool:
    """
    Figure out from /etc/shadow if root is already locked. The output of this
    is stored in the log, so use grep to only log the line for root, not the
    line for the user which contains a hash of the user's password.

    :param suffix: either rootfs_{args.device} or installer_{args.device}
    """
    shadow_root = pmb.chroot.root(
        ["grep", "^root:!:", "/etc/shadow"], chroot, output_return=True, check=False
    )
    return shadow_root.startswith("root:!:")


def setup_login(config: Config, chroot: Chroot, password: str) -> None:
    """
    Loop until the password for user has been set successfully, and disable
    root login.

    :param suffix: of the chroot, where passwd will be execute (either the
                   rootfs_{args.device} or installer_{args.device}
    """
    # User password
    logging.info(f" *** SET LOGIN PASSWORD FOR: '{config.user}' ***")
    setup_login_chpasswd_user_from_arg(config.user, password, chroot)

    # Disable root login
    if is_root_locked(chroot):
        logging.debug(f"({chroot}) root is already locked")
    else:
        logging.debug(f"({chroot}) locking root")
        pmb.chroot.root(["passwd", "-l", "root"], chroot)


def copy_ssh_keys(config: Config) -> None:
    """If requested, copy user's SSH public keys to the device if they exist"""
    if not config.ssh_keys:
        return
    keys = []
    for key in glob.glob(os.path.expanduser(config.ssh_key_glob)):
        with open(key) as infile:
            try:
                keys += infile.readlines()
            except UnicodeDecodeError:
                logging.info(
                    f"WARNING: {key} is not a valid SSH key. "
                    "This file will not be copied to device."
                )

    if not len(keys):
        logging.info(
            "NOTE: Public SSH keys not found. Since no SSH keys "
            "were copied, you will need to use SSH password "
            "authentication!"
        )
        return

    authorized_keys = Chroot.native() / "tmp/authorized_keys"
    with open(authorized_keys, "w") as outfile:
        for key in keys:
            outfile.write(f"{key}")

    target = Chroot.native() / "mnt/install/home/" / config.user / ".ssh"
    pmb.helpers.run.root(["mkdir", target])
    pmb.helpers.run.root(["chmod", "700", target])
    pmb.helpers.run.root(["cp", authorized_keys, target / "authorized_keys"])
    pmb.helpers.run.root(["rm", authorized_keys])
    pmb.helpers.run.root(["chown", "-R", "10000:10000", target])


def setup_keymap(config: Config) -> None:
    """Set the keymap with the setup-keymap utility if the device requires it"""
    chroot = Chroot(ChrootType.ROOTFS, config.device)
    deviceinfo = pmb.parse.deviceinfo(device=config.device)
    if not deviceinfo.keymaps or deviceinfo.keymaps.strip() == "":
        logging.info("NOTE: No valid keymap specified for device")
        return
    options = deviceinfo.keymaps.split(" ")
    if config.keymap != "" and config.keymap in options:
        layout, variant = config.keymap.split("/")
        if pmb.config.is_systemd_selected():
            pmb.chroot.root(
                ["systemd-firstboot", "--force", f"--keymap={variant}"],
                chroot,
                output=RunOutputTypeDefault.LOG,
            )
        else:
            pmb.chroot.root(
                ["setup-keymap", layout, variant], chroot, output=RunOutputTypeDefault.INTERACTIVE
            )

        # Check xorg config
        xconfig = None
        if (chroot / "etc/X11/xorg.conf.d").exists():
            xconfig = pmb.chroot.root(
                ["grep", "-rl", "XkbLayout", "/etc/X11/xorg.conf.d/"],
                chroot,
                check=False,
                output_return=True,
            )
        if xconfig:
            # Nokia n900 (RX-51) randomly merges some keymaps so we
            # have to specify a composite keymap for a few countries. See:
            # https://gitlab.freedesktop.org/xkeyboard-config/xkeyboard-config/-/blob/master/symbols/nokia_vndr/rx-51
            match variant:
                case "rx51_fi" | "rx51_se":
                    layout = "fise"
                case "rx51_da" | "rx51_no":
                    layout = "dano"
                case "rx51_pt" | "rx51_es":
                    layout = "ptes"
                case _:
                    pass
            # Multiple files can contain the keyboard layout, take last
            xconfig = xconfig.splitlines()[-1]
            old_text = 'Option *\\"XkbLayout\\" *\\".*\\"'
            new_text = 'Option \\"XkbLayout\\" \\"' + layout + '\\"'
            pmb.chroot.root(["sed", "-i", "s/" + old_text + "/" + new_text + "/", xconfig], chroot)
    else:
        logging.info("NOTE: No valid keymap specified for device")


def setup_timezone(chroot: Chroot, timezone: str) -> None:
    pmb.chroot.root(["setup-timezone", "-i", timezone], chroot)


def setup_locale(chroot: Chroot, locale: str) -> None:
    """Set locale-related settings such as $LANG and keyboard layout"""
    # 10locale-pmos.sh gets sourced before 20locale.sh from
    # alpine-baselayout by /etc/profile. Since they don't override the
    # locale if it exists, it warranties we have preference
    pmb.chroot.root(["sh", "-c", f"echo LANG={shlex.quote(locale)} > /etc/locale.conf"], chroot)
    # musl-locales doesn't read from /etc/locale.conf, only from environment variables
    # TODO: once musl implements locales (hopefully with /etc/locale.conf support) the following should be removed
    pmb.chroot.root(
        ["sh", "-c", "echo source /etc/locale.conf > /etc/profile.d/10locale-pmos.sh"], chroot
    )
    # add keyboard layout related to locale and layout switcher
    xkb_layout = get_xkb_layout(locale)
    xkb_vars = xkb_layout.get_profile_vars()
    if xkb_vars:
        xkb_vars = xkb_vars.replace("\n", "\\n")
        pmb.chroot.root(
            ["sed", "-i", "$a\\" + xkb_vars, "/etc/profile.d/10locale-pmos.sh"],
            chroot,
        )
    if (chroot / "etc/X11").exists() and (kb_config := xkb_layout.get_keyboard_config()):
        config_name = "99-keyboard.conf"
        config_tmp_path = f"/tmp/{config_name}"
        config_path = f"/etc/X11/xorg.conf.d/{config_name}"
        pmb.chroot.root(["mkdir", "-p", "/etc/X11/xorg.conf.d"], chroot)
        (chroot / config_tmp_path).write_text(kb_config)
        pmb.chroot.root(["mv", config_tmp_path, config_path], chroot)
        pmb.chroot.root(["chown", "root:root", config_path], chroot)


def setup_hostname(device: str, hostname: str | None) -> None:
    """Set the hostname and update localhost address in /etc/hosts"""
    # Default to device name. If device name is not a valid hostname then
    # default to a static default.
    if not hostname:
        hostname = pmb.helpers.other.normalize_hostname(device)
        if not pmb.helpers.other.validate_hostname(hostname):
            # A valid host name, see:
            # https://datatracker.ietf.org/doc/html/rfc1035#section-2.3.1
            hostname = "postmarketos-device"
    elif not pmb.helpers.other.validate_hostname(hostname):
        # Invalid hostname set by the user e.g., via pmb init, this should
        # fail so they can fix it
        raise RuntimeError(
            "Hostname '" + hostname + "' is not valid, please"
            " run 'pmbootstrap init' to configure it."
        )

    suffix = Chroot(ChrootType.ROOTFS, device)
    # Generate /etc/hostname
    pmb.chroot.root(["sh", "-c", "echo " + shlex.quote(hostname) + " > /etc/hostname"], suffix)
    # Update /etc/hosts
    regex = (
        r"s/^127\.0\.0\.1.*/127.0.0.1\t" + re.escape(hostname) + " localhost.localdomain localhost/"
    )
    pmb.chroot.root(["sed", "-i", "-e", regex, "/etc/hosts"], suffix)


def setup_appstream(offline: bool, chroot: Chroot) -> None:
    """
    If alpine-appstream-downloader has been downloaded, execute it to have
    update AppStream data on new installs
    """
    installed_pkgs = pmb.chroot.apk.installed(chroot)

    if "alpine-appstream-downloader" not in installed_pkgs or offline:
        return

    if not pmb.chroot.root(
        ["alpine-appstream-downloader", "/mnt/appstream-data"], chroot, check=False
    ):
        pmb.chroot.root(["mkdir", "-p", "/var/lib/swcatalog"], chroot)
        pmb.chroot.root(
            [
                "cp",
                "-r",
                "/mnt/appstream-data/icons",
                "/mnt/appstream-data/xml",
                "-t",
                "/var/lib/swcatalog",
            ],
            chroot,
        )


def print_sshd_info(no_sshd: bool) -> None:
    logging.info("")  # make the note stand out
    logging.info("*** SSH DAEMON INFORMATION ***")

    if no_sshd:
        logging.info("SSH daemon is disabled (--no-sshd).")
    else:
        logging.info("SSH daemon is enabled (disable with --no-sshd).")
        logging.info(
            f"Login as '{get_context().config.user}' with the password given during installation."
        )


def disable_service_systemd(chroot: Chroot, service_name: str) -> None:
    preset = f"/usr/lib/systemd/system-preset/80-pmbootstrap-install-disable-{service_name}.preset"
    preset_content = f"disable {service_name}.service"
    pmb.chroot.root(
        ["sh", "-c", f"echo {shlex.quote(preset_content)} > {shlex.quote(preset)}"], chroot
    )
    pmb.chroot.root(["systemctl", "preset", f"{service_name}.service"], chroot)

    # Use the output instead of exit code to be more explicit and so we can put
    # it in the error message
    output = pmb.chroot.root(
        ["systemctl", "is-enabled", f"{service_name}.service"],
        chroot,
        output_return=True,
        check=False,
    )
    output = output.rstrip()
    if output != "disabled":
        raise RuntimeError(f"Failed to disable {service_name} service (systemd preset): {output}")


def disable_service_openrc(chroot: Chroot, service_name: str) -> None:
    # check=False: rc-update doesn't exit with 0 if already disabled
    pmb.chroot.root(["rc-update", "del", service_name, "default"], chroot, check=False)

    # Verify that it's gone
    runlevel_files = pmb.helpers.run.root(
        ["find", "-name", service_name], output_return=True, working_dir=chroot / "etc/runlevels"
    )
    if runlevel_files:
        raise RuntimeError(f"Failed to disable service {service_name} (openrc): {runlevel_files}")


def disable_service(chroot: Chroot, service_name: str) -> None:
    if pmb.config.is_systemd_selected():
        disable_service_systemd(chroot, service_name)
    else:
        disable_service_openrc(chroot, service_name)


def print_firewall_info(disabled: bool, arch: Arch) -> None:
    pmaports_cfg = pmb.config.pmaports.read_config()
    pmaports_ok = pmaports_cfg.get("supported_firewall", None) == "nftables"

    # Find kernel pmaport (will not be found if Alpine kernel is used)
    apkbuild_found = False
    apkbuild_has_opt = False

    kernel = get_kernel_package(get_context().config)
    if kernel:
        _, kernel_apkbuild = pmb.build.get_apkbuild(kernel[0])
        if kernel_apkbuild:
            opts = kernel_apkbuild["options"]
            apkbuild_has_opt = "pmb:kconfigcheck-nftables" in opts
            apkbuild_found = True

    # Print the note and make it stand out
    logging.info("")
    logging.info("*** FIREWALL INFORMATION ***")

    if not pmaports_ok:
        logging.info("Firewall is not supported in checked out pmaports branch.")
    elif disabled:
        logging.info("Firewall is disabled (--no-firewall).")
    elif not apkbuild_found:
        logging.info(
            "Firewall is enabled, but may not work (couldn't"
            " determine if kernel supports nftables)."
        )
    elif apkbuild_has_opt:
        logging.info("Firewall is enabled and supported by kernel.")
    else:
        logging.info(
            "Firewall is enabled, but will not work (no support in kernel config for nftables)."
        )
        logging.info("If/when the kernel supports it in the future, it will work automatically.")

    logging.info("For more information: https://postmarketos.org/firewall")


def generate_binary_list(chroot: Chroot, step: int) -> list[tuple[str, int]]:
    """
    Perform three checks prior to writing binaries to disk: 1) that binaries
    exist, 2) that binaries do not extend into the first partition, 3) that
    binaries do not overlap each other.

    :param suffix: of the chroot, which holds the firmware files (either the
                   rootfs_{args.device} or installer_{args.device}
    :param step: partition step size in bytes
    """
    binary_ranges: dict[int, int] = {}
    binary_list = []
    binaries = (pmb.parse.deviceinfo().sd_embed_firmware or "").split(",")

    for binary_offset in binaries:
        binary, offset_ = binary_offset.split(":")
        try:
            offset = int(offset_)
        except ValueError as exception:
            raise RuntimeError(
                f"Value for firmware binary offset is not valid: {offset_}"
            ) from exception
        binary_path = chroot / "usr/share" / binary
        if not os.path.exists(binary_path):
            raise RuntimeError(
                "The following firmware binary does not "
                f"exist in the {chroot} chroot: "
                f"/usr/share/{binary}"
            )
        # Insure that embedding the firmware will not overrun the
        # first partition
        boot_part_start = pmb.parse.deviceinfo().boot_part_start or "2048"
        max_size = (int(boot_part_start) * 512) - (offset * step)
        binary_size = os.path.getsize(binary_path)
        if binary_size > max_size:
            raise RuntimeError(
                f"The firmware is too big to embed in the disk image {binary_size}B > {max_size}B"
            )
        # Insure that the firmware does not conflict with any other firmware
        # that will be embedded
        binary_start = offset * step
        binary_end = binary_start + binary_size
        for start, end in binary_ranges.items():
            if (binary_start >= start and binary_start < end) or (
                binary_end > start and binary_end <= end
            ):
                raise RuntimeError(
                    f"The firmware overlaps with at least one other firmware image: {binary}"
                )

        binary_ranges[binary_start] = binary_end
        binary_list.append((binary, offset))

    return binary_list


def embed_firmware(suffix: Chroot) -> None:
    """
    This method will embed firmware, located at /usr/share, that are specified
    by the "sd_embed_firmware" deviceinfo parameter into the SD card image
    (e.g. u-boot). Binaries that would overwrite the first partition are not
    accepted, and if multiple binaries are specified then they will be checked
    for collisions with each other.

    :param suffix: of the chroot, which holds the firmware files (either the
                   rootfs_{args.device} or installer_{args.device}
    """
    if not pmb.parse.deviceinfo().sd_embed_firmware:
        return

    step = 1024
    if pmb.parse.deviceinfo().sd_embed_firmware_step_size:
        try:
            step = int(pmb.parse.deviceinfo().sd_embed_firmware_step_size or "invalid")
        except ValueError as exception:
            raise RuntimeError(
                f"Value for deviceinfo_sd_embed_firmware_step_size is not valid: {step}"
            ) from exception

    device_rootfs = mount_device_rootfs(suffix)
    binary_list = generate_binary_list(suffix, step)

    # Write binaries to disk
    for binary, offset in binary_list:
        binary_file = os.path.join("/usr/share", binary)
        logging.info(
            f"Embed firmware {binary} in the SD card image at offset {offset} with step size {step}"
        )
        filename = os.path.join(device_rootfs, binary_file.lstrip("/"))
        pmb.chroot.root(
            ["dd", "if=" + filename, "of=/dev/install", "bs=" + str(step), "seek=" + str(offset)]
        )


def write_cgpt_kpart(layout: PartitionLayout, suffix: Chroot, install_cgpt: bool) -> None:
    """
    Write the kernel to the ChromeOS kernel partition.

    :param layout: partition layout from get_partition_layout()
    :param suffix: of the chroot, which holds the image file to be flashed
    """
    if not pmb.parse.deviceinfo().cgpt_kpart or not install_cgpt:
        return

    device_rootfs = mount_device_rootfs(suffix)
    filename = f"{device_rootfs}{pmb.parse.deviceinfo().cgpt_kpart}"
    pmb.chroot.root(["dd", f"if={filename}", f"of=/dev/installp{layout['kernel']}"])


def sanity_check_boot_size() -> None:
    default = Config().boot_size
    config = get_context().config
    if int(config.boot_size) >= int(default):
        return
    logging.error(
        "ERROR: your pmbootstrap has a small/invalid boot_size of"
        f" {config.boot_size} configured, probably because the config"
        " has been created with an old version."
    )
    logging.error(f"This can lead to problems later on, we recommend setting it to {default} MiB.")
    logging.error(f"Run 'pmbootstrap config boot_size {default}' and try again.")
    sys.exit(1)


def sanity_check_disk(device: Path) -> None:
    device_name = os.path.basename(device)
    if not os.path.exists(device):
        raise RuntimeError(f"{device} doesn't exist, is the disk plugged?")
    if os.path.isdir(f"/sys/class/block/{device_name}"):
        ro = Path(f"/sys/class/block/{device_name}/ro").read_text()
        if ro == "1\n":
            raise RuntimeError(f"{device} is read-only, maybe a locked SD card?")


def sanity_check_disk_size(device: Path) -> None:
    devpath = os.path.realpath(device)
    sysfs = "/sys/class/block/{}/size".format(devpath.replace("/dev/", ""))
    if not os.path.isfile(sysfs):
        # This is a best-effort sanity check, continue if it's not checkable
        return

    raw = Path(sysfs).read_text()

    # Size is in 512-byte blocks
    size = int(raw.strip())
    human = f"{size / 2 / 1024 / 1024:.2f} GiB"

    # Warn if the size is larger than 100GiB
    if (
        not get_context().assume_yes
        and size > (100 * 2 * 1024 * 1024)
        and not pmb.helpers.cli.confirm(
            f"WARNING: The target disk ({devpath}) "
            "is larger than a usual SD card "
            "(>100GiB). Are you sure you want to "
            f"overwrite this {human} disk?",
            no_assumptions=True,
        )
    ):
        raise RuntimeError("Aborted.")


def get_partition_layout(kernel: bool, prep: bool | None) -> PartitionLayout:
    """
    :param kernel: create a separate kernel partition before all other
                   partitions, e.g. for the ChromeOS devices with cgpt
    :param prep: create a PReP Boot partition before root and no boot partition,
                 e.g. for PowerVM or OpenFirmware POWER machines
    :returns: the partition layout, e.g. without reserve and kernel:
              {"kernel": None, "boot": 1, "reserve": None, "root": 2}
    """
    ret: PartitionLayout = {
        "kernel": None,
        "prep": None,
        "boot": 1,
        "reserve": None,
        "root": 2,
    }

    if prep:
        ret["prep"] = 1
        ret["boot"] = None
    else:
        if kernel:
            ret["kernel"] = 1
            ret["boot"] = 2
            ret["root"] += 1

    return ret


def get_uuid(partition: Path) -> str:
    """
    Get UUID of a partition

    :param partition: block device for getting UUID from
    """
    return pmb.chroot.root(
        [
            "blkid",
            "-s",
            "UUID",
            "-o",
            "value",
            partition,
        ],
        output_return=True,
    ).rstrip()


def create_crypttab(layout: PartitionLayout | None, chroot: Chroot) -> None:
    """
    Create /etc/crypttab config

    :param layout: partition layout from get_partition_layout() or None
    :param suffix: of the chroot, which crypttab will be created to
    """
    root_dev = Path(f"/dev/installp{layout['root']}") if layout else Path("/dev/install")

    luks_uuid = get_uuid(root_dev)
    crypttab = f"root UUID={luks_uuid} none luks\n"

    (chroot / "tmp/crypttab").open("w").write(crypttab)
    pmb.chroot.root(["mv", "/tmp/crypttab", "/etc/crypttab"], chroot)


def create_fstab(
    layout: PartitionLayout | None,
    chroot: Chroot,
    filesystem: str,
    full_disk_encryption: bool,
) -> None:
    """
    Create /etc/fstab config

    :param layout: partition layout from get_partition_layout() or None
    :param chroot: of the chroot, which fstab will be created to
    """
    uses_prep = layout and layout["prep"] is not None

    if layout and not uses_prep:
        boot_dev = Path(f"/dev/installp{layout['boot']}")
        root_dev = Path(f"/dev/installp{layout['root']}")
    elif layout and uses_prep:
        boot_dev = None
        root_dev = Path(f"/dev/installp{layout['root']}")
    else:
        boot_dev = None
        root_dev = Path("/dev/install")

    root_mount_point = "/dev/mapper/root" if full_disk_encryption else f"UUID={get_uuid(root_dev)}"
    root_filesystem = pmb.install.get_root_filesystem(filesystem)

    if root_filesystem == "btrfs":
        # btrfs gets separate subvolumes for root, var and home
        fstab = f"""
# <file system> <mount point> <type> <options> <dump> <pass>
{root_mount_point} / btrfs subvol=@,compress=zstd:2,ssd 0 0
{root_mount_point} /home btrfs subvol=@home,compress=zstd:2,ssd 0 0
{root_mount_point} /root btrfs subvol=@root,compress=zstd:2,ssd 0 0
{root_mount_point} /srv btrfs subvol=@srv,compress=zstd:2,ssd 0 0
{root_mount_point} /var btrfs subvol=@var,ssd 0 0
{root_mount_point} /.snapshots btrfs subvol=@snapshots,compress=zstd:2,ssd 0 0
""".lstrip()

    else:
        fstab = f"""
# <file system> <mount point> <type> <options> <dump> <pass>
{root_mount_point} / {root_filesystem} defaults 0 0
""".lstrip()

    if boot_dev:
        boot_mount_point = f"UUID={get_uuid(boot_dev)}"
        boot_options = "nodev,nosuid,noexec"
        boot_filesystem = pmb.parse.deviceinfo().boot_filesystem or "ext2"
        if boot_filesystem in ("fat16", "fat32"):
            boot_filesystem = "vfat"
            boot_options += ",umask=0077,nosymfollow,codepage=437,iocharset=ascii"
        fstab += f"{boot_mount_point} /boot {boot_filesystem} {boot_options} 0 0\n"

    with (chroot / "tmp/fstab").open("w") as f:
        f.write(fstab)
    pmb.chroot.root(["mv", "/tmp/fstab", "/etc/fstab"], chroot)


def install_system_image(
    args: PmbArgs,
    chroot: Chroot,
    step: int,
    steps: int,
    boot_label: str = "pmOS_boot",
    root_label: str = "pmOS_root",
    split: bool = False,
    single_partition: bool = False,
    disk: Path | None = None,
) -> None:
    """
    :param size_reserve: empty partition between root and boot in MiB (pma#463)
    :param suffix: the chroot suffix, where the rootfs that will be installed
                   on the device has been created (e.g. "rootfs_qemu-amd64")
    :param step: next installation step
    :param steps: total installation steps
    :param boot_label: label of the boot partition (e.g. "pmOS_boot")
    :param root_label: label of the root partition (e.g. "pmOS_root")
    :param split: create separate images for boot and root partitions
    :param disk: path to disk block device (e.g. /dev/mmcblk0) or None
    """
    config = get_context().config
    device = chroot.name
    # Partition and fill image file/disk block device
    logging.info(f"*** ({step}/{steps}) PREPARE INSTALL BLOCKDEVICE ***")
    pmb.helpers.mount.umount_all(chroot.path)
    (size_boot, size_root) = get_subpartitions_size(chroot)
    if not single_partition:
        layout = get_partition_layout(
            bool(pmb.parse.deviceinfo().cgpt_kpart) and args.install_cgpt,
            pmb.parse.deviceinfo().create_prep_boot,
        )
    else:
        layout = None
    if not args.rsync:
        # If disk is set, we already handled this earlier.
        if not disk:
            pmb.install.blockdevice.create(args.sector_size, size_boot, size_root, split)
        if not split and layout:
            if pmb.parse.deviceinfo().cgpt_kpart and args.install_cgpt:
                pmb.install.partition_cgpt(layout, size_boot)
            elif pmb.parse.deviceinfo().create_prep_boot:
                pmb.install.partition_prep(layout)
            else:
                pmb.install.partition(layout, size_boot)

    # Inform kernel about changed partition table in case parted couldn't
    pmb.chroot.root(["partprobe", "/dev/install"], check=False)

    if not split and not single_partition:
        assert layout  # Initialized above for not single_partition case (mypy needs this)
        pmb.install.partitions_mount(device, layout, disk)

    pmb.install.format(
        layout,
        boot_label,
        root_label,
        disk,
        args.rsync,
        args.filesystem,
        args.full_disk_encryption,
        args.cipher,
        args.iter_time,
    )

    # Since we shut down the chroot we need to mount it again
    pmb.chroot.mount(chroot)

    # Create /etc/fstab and /etc/crypttab
    logging.info("(native) create /etc/fstab")
    create_fstab(layout, chroot, args.filesystem, args.full_disk_encryption)
    if args.full_disk_encryption:
        logging.info("(native) create /etc/crypttab")
        create_crypttab(layout, chroot)

    # Run mkinitfs to pass UUIDs to cmdline
    logging.info(f"({chroot}) mkinitfs")
    pmb.chroot.root(["mkinitfs"], chroot)

    # Clean up after running mkinitfs in chroot
    pmb.helpers.mount.umount_all(chroot.path)
    pmb.helpers.run.root(["rm", chroot / "in-pmbootstrap"])
    pmb.chroot.remove_mnt_pmbootstrap(chroot)

    # Just copy all the files
    logging.info(f"*** ({step + 1}/{steps}) FILL INSTALL BLOCKDEVICE ***")
    copy_files_from_chroot(chroot, args.rsync, args.verbose)
    create_home_from_skel(args.filesystem, config.user)
    configure_apk(args.install_local_pkgs)
    copy_ssh_keys(config)

    # Don't try to embed firmware and cgpt on split images since there's no
    # place to put it and it will end up in /dev of the chroot instead
    if not split and not single_partition:
        assert layout  # Initialized above for not single_partition case (mypy needs this)
        embed_firmware(chroot)
        write_cgpt_kpart(layout, chroot, args.install_cgpt)

    # Install GRUB to the PReP boot partition
    if pmb.parse.deviceinfo().create_prep_boot and layout:
        pmb.chroot.apk.install(["grub"], Chroot.native())
        # GRUB is incapable of installing to /dev/installp1, it *needs* to install to
        # /dev/loop0p1
        pmb.chroot.root(
            [
                "grub-install",
                "--target=powerpc-ieee1275",
                "--directory=/mnt/install/usr/lib/grub/powerpc-ieee1275",
                "--boot-directory=/mnt/install/boot",
                # Since we are in a chroot, it cannot properly autodetect the modules,
                # so we need to give it a hint
                "--modules=part_gpt part_msdos disk normal linux boot search terminal",
                "--no-nvram",
                f"/dev/loop0p{layout['prep']}",
            ]
        )

    if disk:
        logging.info(f"Unmounting disk {disk} (this may take a while to sync, please wait)")
    pmb.chroot.shutdown(True)

    # Convert rootfs to sparse using img2simg
    sparse = args.sparse
    if sparse is None:
        sparse = pmb.parse.deviceinfo().flash_sparse == "true"

    if sparse and not split and not disk:
        workdir = Path("/home/pmos/rootfs")
        logging.info("(native) make sparse rootfs")
        pmb.chroot.apk.install(["android-tools"], Chroot.native())
        sys_image = device + ".img"
        sys_image_sparse = device + "-sparse.img"
        pmb.chroot.user(["img2simg", sys_image, sys_image_sparse], working_dir=workdir)
        pmb.chroot.user(["mv", "-f", sys_image_sparse, sys_image], working_dir=workdir)

        # patch sparse image for Samsung devices if specified
        samsungify_strategy = pmb.parse.deviceinfo().flash_sparse_samsung_format
        if samsungify_strategy:
            logging.info("(native) convert sparse image into Samsung's sparse image format")
            pmb.chroot.apk.install(["sm-sparse-image-tool"], Chroot.native())
            sys_image = f"{device}.img"
            sys_image_patched = f"{device}-patched.img"
            pmb.chroot.user(
                [
                    "sm_sparse_image_tool",
                    "samsungify",
                    "--strategy",
                    samsungify_strategy,
                    sys_image,
                    sys_image_patched,
                ],
                working_dir=workdir,
            )
            pmb.chroot.user(["mv", "-f", sys_image_patched, sys_image], working_dir=workdir)


def print_flash_info(
    device: str, deviceinfo: Deviceinfo, split: bool, have_disk: bool, single_partition: bool
) -> None:
    """
    Print flashing information, based on the deviceinfo data and the
    pmbootstrap arguments.
    """
    logging.info("")  # make the note stand out
    logging.info("*** FLASHING INFORMATION ***")

    # System flash information
    method = deviceinfo.flash_method
    flasher = pmb.config.flashers.get(method, {})
    flasher_actions = flasher.get("actions", {})
    if not isinstance(flasher_actions, dict):
        raise TypeError(f"flasher actions must be a dictionary, got: {flasher_actions}")
    requires_split = flasher.get("split", False)

    if method == "none":
        logging.info(
            "Refer to the installation instructions of your device,"
            " or the generic install instructions in the wiki."
        )
        logging.info("https://wiki.postmarketos.org/wiki/Installation_guide#pmbootstrap_flash")
        return

    logging.info("Run the following to flash your installation to the target device:")

    if "flash_rootfs" in flasher_actions and not have_disk and bool(split) == requires_split:
        logging.info("* pmbootstrap flasher flash_rootfs")
        logging.info("  Flashes the generated rootfs image to your device:")
        if split:
            logging.info(f"  {Chroot.native() / 'home/pmos/rootfs' / device}-root.img")
        else:
            logging.info(f"  {Chroot.native() / 'home/pmos/rootfs' / device}.img")
            if not single_partition:
                logging.info(
                    "  (NOTE: This file has a partition table, which"
                    " contains /boot and / subpartitions. That way we"
                    " don't need to change the partition layout on your"
                    " device.)"
                )

    # if current flasher supports vbmeta and partition is explicitly specified
    # in deviceinfo
    if "flash_vbmeta" in flasher_actions and (
        deviceinfo.flash_fastboot_partition_vbmeta or deviceinfo.flash_heimdall_partition_vbmeta
    ):
        logging.info("* pmbootstrap flasher flash_vbmeta")
        logging.info("  Flashes vbmeta image with verification disabled flag.")

    # if current flasher supports dtbo and partition is explicitly specified
    # in deviceinfo
    if "flash_dtbo" in flasher_actions and (
        deviceinfo.flash_fastboot_partition_dtbo or deviceinfo.flash_heimdall_partition_dtbo
    ):
        logging.info("* pmbootstrap flasher flash_dtbo")
        logging.info("  Flashes dtbo image.")

    # Most flash methods operate independently of the boot partition.
    # (e.g. an Android boot image is generated). In that case, "flash_kernel"
    # works even when partitions are split or installing to disk. This is not
    # possible if the flash method requires split partitions.
    if "flash_kernel" in flasher_actions and (not requires_split or split):
        logging.info("* pmbootstrap flasher flash_kernel")
        logging.info("  Flashes the kernel + initramfs to your device:")
        if requires_split:
            logging.info(f"  {Chroot.native() / 'home/pmos/rootfs' / device}-boot.img")
        else:
            logging.info(f"  {Chroot(ChrootType.ROOTFS, device) / 'boot'}")

    if "flash_boot" in flasher_actions and (Chroot.rootfs(device) / "boot/boot.img").exists():
        logging.info("* pmbootstrap flasher flash_boot")
        logging.info("  Flashes the generated Android boot image to your device:")
        logging.info(f"  {Chroot.rootfs(device) / 'boot/boot.img'}")
        logging.info("  (NOTE: This is not necessary if using a custom bootloader like U-Boot)")

    if "boot" in flasher_actions or "boot_gki" in flasher_actions:
        logging.info(
            "  (NOTE: " + method + " also supports booting"
            " the kernel/initramfs directly without flashing."
            " Use 'pmbootstrap flasher boot' to do that.)"
        )

    if (
        "flash_lk2nd" in flasher_actions
        and (Chroot(ChrootType.ROOTFS, device) / "boot/lk2nd.img").exists()
    ):
        logging.info(
            "* Your device supports and may even require"
            " flashing lk2nd. You should flash it before"
            " flashing anything else. Use 'pmbootstrap flasher"
            " flash_lk2nd' to do that."
        )

    # Export information
    logging.info(
        "* If the above steps do not work, you can also create"
        " symlinks to the generated files with 'pmbootstrap export'"
        " and flash outside of pmbootstrap."
    )


def install_recovery_zip(args: PmbArgs, device: str, arch: Arch, steps: int) -> None:
    logging.info(f"*** ({steps}/{steps}) CREATING RECOVERY-FLASHABLE ZIP ***")
    chroot = Chroot(ChrootType.BUILDROOT, arch)
    pmb.chroot.init(chroot)
    mount_device_rootfs(Chroot.rootfs(device), chroot)
    pmb.install.recovery.create_zip(args, chroot, device)

    # Flash information
    logging.info("*** FLASHING INFORMATION ***")
    logging.info("Flashing with the recovery zip is explained here:")
    logging.info("https://postmarketos.org/recoveryzip")


def get_selected_providers(packages: list[str]) -> list[str]:
    """
    Look through the specified packages and see which providers were selected
    in "pmbootstrap init". Install those as extra packages to select them
    instead of the default provider. This function is called recursively on the
    dependencies of the given packages.

    :param packages: the packages that have selectable providers (_pmb_select)
    :param initial: used internally when the function calls itself
    :return: additional provider packages to install
    """
    global get_selected_providers_visited

    ret = []

    for package in packages:
        if package in get_selected_providers_visited:
            logging.verbose(f"get_selected_providers: {package}: already visited")
            continue
        get_selected_providers_visited += [package]

        # Note that this ignores packages that don't exist. This means they
        # aren't in pmaports. This is fine, with the assumption that
        # installation will fail later in some other method if they truly don't
        # exist in any repo.
        apkbuild = pmb.helpers.pmaports.get(package, subpackages=False, must_exist=False)
        if not apkbuild:
            continue
        for select in apkbuild["_pmb_select"]:
            if select in get_context().config.providers:
                ret += [get_context().config.providers[select]]
                logging.verbose(f"{package}: install selected_providers: {', '.join(ret)}")
            else:
                for default in apkbuild["_pmb_default"]:
                    # default: e.g. "postmarketos-base-ui-audio-pipewire"
                    # select: e.g. "postmarketos-base-ui-audio"
                    if default.startswith(f"{select}-"):
                        ret += [default]
        # Also iterate through dependencies to collect any providers they have
        depends = apkbuild["depends"]
        if depends:
            ret += get_selected_providers(depends)

    return ret


def get_recommends(packages: list[str], install_recommends: bool) -> Sequence[str]:
    """
    Look through the specified packages and collect additional packages
    specified under _pmb_recommends in them. This is recursive, so it will dive
    into packages that are listed under recommends to collect any packages they
    might also have listed under their own _pmb_recommends.

    Recursion is only done into packages found in pmaports.

    If running with pmbootstrap install --no-recommends, this function returns
    an empty list.

    :param packages: list of packages of which we want to get the recommends
    :param initial: used internally when the function calls itself
    :returns: list of pkgnames, e.g. ["chatty", "gnome-contacts"]
    """
    global get_recommends_visited

    ret: list[str] = []
    if not install_recommends:
        return ret

    for package in packages:
        if package in get_recommends_visited:
            logging.verbose(f"get_recommends: {package}: already visited")
            continue
        get_recommends_visited += [package]

        # Note that this ignores packages that don't exist. This means they
        # aren't in pmaports. This is fine, with the assumption that
        # installation will fail later in some other method if they truly don't
        # exist in any repo.
        apkbuild = pmb.helpers.pmaports.get(package, must_exist=False)
        if not apkbuild:
            continue
        if package in apkbuild["subpackages"]:
            # Just focus on the subpackage
            apkbuild = apkbuild["subpackages"][package]
            # The subpackage is None if the subpackage does not have a function
            # in the APKBUILD (uses the default function), e.g. for most openrc
            # subpackages. See pmb.parse._apkbuild._parse_subpackage().
            if not apkbuild:
                continue
        recommends = apkbuild["_pmb_recommends"]
        if recommends:
            logging.debug(f"{package}: install _pmb_recommends: {', '.join(recommends)}")
            ret += recommends
            # Call recursively in case recommends have pmb_recommends of their
            # own.
            ret += get_recommends(recommends, install_recommends)
        # Also iterate through dependencies to collect any recommends they have
        depends = apkbuild["depends"]
        if depends:
            ret += get_recommends(depends, install_recommends)

    return ret


def create_device_rootfs(
    step: int,
    steps: int,
    add: str,
    password: str,
    install_base: bool,
    install_recommends: bool,
    full_disk_encryption: bool,
    no_sshd: bool,
    no_firewall: bool,
) -> None:
    # list all packages to be installed (including the ones specified by --add)
    # and upgrade the installed packages/apkindexes
    context = get_context()
    config = context.config
    device = context.config.device
    logging.info(f'*** ({step}/{steps}) CREATE DEVICE ROOTFS ("{device}") ***')

    chroot = Chroot(ChrootType.ROOTFS, device)
    pmb.chroot.init(chroot)
    # Create user before installing packages, so post-install scripts of
    # pmaports can figure out the username (legacy reasons: pmaports#820)
    set_user(context.config)

    # Fill install_packages
    install_packages = [*pmb.config.install_device_packages, "device-" + device]
    if not install_base:
        install_packages = [p for p in install_packages if p != "postmarketos-base"]
    ui_package_name = f"postmarketos-ui-{config.ui}"
    if config.ui.lower() != "none":
        install_packages += [ui_package_name]

    # Add additional providers of base/device/UI package
    install_packages += get_selected_providers(install_packages)

    install_packages += get_kernel_package(config)
    install_packages += get_nonfree_packages(device)
    if context.config.ui.lower() != "none":
        ui_package = pmb.helpers.pmaports.get(ui_package_name, subpackages=False, must_exist=False)
        if ui_package and context.config.ui_extras:
            extra = f"postmarketos-ui-{config.ui}-extras"
            extra_package = ui_package["subpackages"].get(extra)
            if extra_package:
                install_packages += [extra]

    if context.config.extra_packages.lower() != "none":
        install_packages += context.config.extra_packages.split(",")
    if add:
        install_packages += add.split(",")

    pmaports_cfg = pmb.config.pmaports.read_config()
    # postmarketos-base supports a dummy package for blocking unl0kr install
    # when not required
    if pmaports_cfg.get("supported_base_nofde", None):
        if full_disk_encryption:
            # Pick the most suitable unlocker depending on the packages
            # selected for installation
            unlocker = pmb.parse.depends.package_provider(
                "postmarketos-fde-unlocker", install_packages, chroot
            )
            if not unlocker:
                raise RuntimeError(
                    "Full disk encryption enabled but unable to find any suitable FDE unlocker app"
                )
            if unlocker.pkgname not in install_packages:
                install_packages += [unlocker.pkgname]
        else:
            install_packages += ["postmarketos-base-nofde"]

    pmb.helpers.repo.update(pmb.parse.deviceinfo().arch)

    # Install uninstallable "dependencies" by default
    install_packages += get_recommends(install_packages, install_recommends)

    # Install the base-systemd package first to make sure presets are available
    # when services are installed later
    if pmb.config.other.is_systemd_selected(context.config):
        pmb.chroot.apk.install(["postmarketos-base-systemd"], chroot)

    # Install all packages to device rootfs chroot (and rebuild the initramfs,
    # because that doesn't always happen automatically yet, e.g. when the user
    # installed a hook without pmbootstrap - see #69 for more info)
    # Packages will be built if necessary as part of this step
    pmb.chroot.apk.install(install_packages, chroot)
    pmb.chroot.initfs.build(chroot)

    # Set the user password
    setup_login(config, chroot, password)

    # Set the keymap if the device requires it
    setup_keymap(config)

    # Set timezone
    setup_timezone(chroot, config.timezone)

    # Set locale
    setup_locale(chroot, config.locale)

    # Set the hostname as the device name
    setup_hostname(device, config.hostname)

    setup_appstream(context.offline, chroot)

    if no_sshd:
        disable_service(chroot, "sshd")
    if no_firewall:
        disable_service(chroot, "nftables")


def install(args: PmbArgs, is_split: bool) -> None:
    device = get_context().config.device
    chroot = Chroot(ChrootType.ROOTFS, device)
    deviceinfo = pmb.parse.deviceinfo()
    # Sanity checks
    sanity_check_boot_size()
    if not args.android_recovery_zip and args.disk:
        sanity_check_disk(args.disk)
        sanity_check_disk_size(args.disk)

    # --single-partition implies --no-split. There is nothing to split if
    # there is only a single partition.
    if args.single_partition:
        is_split = False
        if deviceinfo.create_initfs_extra:
            raise RuntimeError("--single-partition does not work for devices with initramfs-extra")

    # Number of steps for the different installation methods.
    if args.no_image:
        steps = 2
    elif args.android_recovery_zip:
        steps = 3
    else:
        steps = 4

    if args.zap:
        pmb.chroot.zap(False)

    # Install required programs in native chroot
    step = 1
    logging.info(f"*** ({step}/{steps}) PREPARE NATIVE CHROOT ***")
    pmb.chroot.init(Chroot.native())
    if args.disk:
        pmb.install.blockdevice.handle_disk_mount(args.disk)
    pmb.chroot.apk.install(pmb.config.install_native_packages, Chroot.native(), build=False)
    step += 1

    create_device_rootfs(
        step,
        steps,
        args.add,
        args.password,
        args.install_base,
        args.install_recommends,
        args.full_disk_encryption,
        args.no_sshd,
        args.no_firewall,
    )
    step += 1

    if args.no_image:
        return
    elif args.android_recovery_zip:
        return install_recovery_zip(args, device, deviceinfo.arch, steps)

    install_system_image(
        args,
        chroot,
        step,
        steps,
        split=is_split,
        single_partition=args.single_partition,
        disk=args.disk,
    )

    print_flash_info(
        device,
        deviceinfo,
        is_split,
        bool(args.disk and args.disk.is_absolute()),
        args.single_partition,
    )
    print_sshd_info(args.no_sshd)
    print_firewall_info(args.no_firewall, deviceinfo.arch)

    # Leave space before 'chroot still active' note
    logging.info("")



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Functions that work with pmaports.

See also:
- pmb/helpers/repo.py (work with binary package repos)
- pmb/helpers/package.py (work with both)
"""

from pathlib import Path
from typing import Any, Literal, overload

import pmb.parse
from pmb.core.arch import Arch
from pmb.core.pkgrepo import pkgrepo_iter_package_dirs
from pmb.helpers import logging
from pmb.meta import Cache
from pmb.types import Apkbuild, WithExtraRepos


@Cache("with_extra_repos")
def _find_apkbuilds(with_extra_repos: WithExtraRepos = "default") -> dict[str, Path]:
    apkbuilds = {}
    for package in pkgrepo_iter_package_dirs(with_extra_repos=with_extra_repos):
        pkgname = package.name
        if pkgname in apkbuilds:
            raise RuntimeError(
                f"Package {pkgname} found in multiple aports subfolders. Please put it only in one folder."
            )
        apkbuilds[pkgname] = package / "APKBUILD"

    # Sort dictionary so we don't need to do it over and over again in
    # get_list()
    return dict(sorted(apkbuilds.items()))


def get_list() -> list[str]:
    """:returns: list of all pmaport pkgnames (["hello-world", ...])"""
    return list(_find_apkbuilds().keys())


def guess_main_dev(subpkgname: str) -> Path | None:
    """
    Check if a package without "-dev" at the end exists in pmaports or not, and log the appropriate message.

    Don't call this function directly, use guess_main() instead.

    :param subpkgname: subpackage name, must end in "-dev"
    :returns: full path to the pmaport or None
    """
    pkgname = subpkgname[:-4]
    path = _find_apkbuilds().get(pkgname)
    if path:
        logging.verbose(
            f"{subpkgname}: guessed to be a subpackage of {pkgname} (just removed '-dev')"
        )
        return path.parent

    logging.verbose(
        f"{subpkgname}: guessed to be a subpackage of {pkgname}, which we can't find in pmaports, so it's probably in Alpine"
    )
    return None


def guess_main_cross(subpkgname: str) -> Path | None:
    """
    Check if a subpackage that is part of the cross toolchain is in pmaports or not, and log the appropriate message.

    Don't call this function directly, use guess_main() instead.

    :param subpkgname: subpackage name
    :returns: full path to the pmaport or None
    """
    # If it contains -dev-, assume the parent package is the same, without the infix
    if "-dev-" in subpkgname:
        pkgname = subpkgname.replace("-dev-", "-")
    else:
        pkgname = subpkgname.replace("g++", "gcc")

    path = _find_apkbuilds().get(pkgname)
    if path:
        logging.verbose(f"{subpkgname}: guessed to be a subpackage of {pkgname}")
        return path.parent

    logging.verbose(
        f"{subpkgname}: guessed to be a subpackage of {pkgname}, which we can't find in pmaports, so it's probably in Alpine"
    )
    return None


def guess_main(subpkgname: str) -> Path | None:
    """
    Find the main package by assuming it is a prefix of the subpkgname.

    We do that, because in some APKBUILDs the subpkgname="" variable gets
    filled with a shell loop and the APKBUILD parser in pmbootstrap can't
    parse this right. (Intentionally, we don't want to implement a full shell
    parser.)

    :param subpkgname: subpackage name (e.g. "u-boot-some-device")
    :returns: * full path to the aport, e.g.:
                "/home/user/code/pmbootstrap/aports/main/u-boot"
              * None when we couldn't find a main package
    """
    # Packages ending in -dev: just assume that the originating aport has the
    # same pkgname, except for the -dev at the end. If we use the other method
    # below on subpackages, we may end up with the wrong package. For example,
    # if something depends on plasma-framework-dev, and plasma-framework is in
    # Alpine, but plasma is in pmaports, then the cutting algorithm below would
    # pick plasma instead of plasma-framework.
    if subpkgname.endswith("-dev"):
        return guess_main_dev(subpkgname)

    # cross/* packages have a bunch of subpackages that do not have the main
    # package name as a prefix (i.e. g++-*). Further, the -dev check fails here
    # since the name ends with the name of the architecture.
    if any(subpkgname.endswith("-" + str(arch)) for arch in Arch.supported()):
        return guess_main_cross(subpkgname)

    # Iterate until the cut up subpkgname is gone
    words = subpkgname.split("-")
    while len(words) > 1:
        # Remove one dash-separated word at a time ("a-b-c" -> "a-b")
        words.pop()
        pkgname = "-".join(words)

        # Look in pmaports
        path = _find_apkbuilds().get(pkgname)
        if path:
            logging.verbose(f"{subpkgname}: guessed to be a subpackage of {pkgname}")
            return path.parent

    return None


def _find_package_in_apkbuild(package: str, path: Path) -> bool:
    """
    Look through subpackages and all provides to see if the APKBUILD at the specified path
    contains (or provides) the specified package.

    :param package: The package to search for
    :param path: The path to the apkbuild
    :return: True if the APKBUILD contains or provides the package
    """
    apkbuild = pmb.parse.apkbuild(path)

    # Subpackages
    if package in apkbuild["subpackages"]:
        return True

    # Search for provides in both package and subpackages
    apkbuild_pkgs = [apkbuild, *apkbuild["subpackages"].values()]
    for apkbuild_pkg in apkbuild_pkgs:
        if not apkbuild_pkg:
            continue

        # Provides (cut off before equals sign for entries like
        # "mkbootimg=0.0.1")
        for provides_i in apkbuild_pkg["provides"]:
            # Ignore provides without version, they shall never be
            # automatically selected
            if "=" not in provides_i:
                continue

            if package == provides_i.split("=", 1)[0]:
                return True

    return False


def show_pkg_not_found_systemd_hint(package: str, with_extra_repos: WithExtraRepos) -> None:
    """
    Check if a package would be found if systemd was enabled and display a
    hint about it.
    """
    if with_extra_repos != "default" or pmb.config.other.is_systemd_selected():
        return

    if find(package, False, with_extra_repos="enabled"):
        logging.info(
            f"NOTE: The package '{package}' exists in extra-repos/systemd, but systemd is currently disabled"
        )


@overload
def find(
    package: str,
    must_exist: Literal[True] = ...,
    subpackages: bool = ...,
    with_extra_repos: WithExtraRepos = ...,
) -> Path: ...


@overload
def find(
    package: str,
    must_exist: bool = ...,
    subpackages: bool = ...,
    with_extra_repos: WithExtraRepos = ...,
) -> Path | None: ...


@Cache("package", "must_exist", "subpackages", "with_extra_repos")
def find(
    package: str,
    must_exist: bool = True,
    subpackages: bool = True,
    with_extra_repos: WithExtraRepos = "default",
) -> Path | None:
    """
    Find the directory in pmaports that provides a package or subpackage.
    If you want the parsed APKBUILD instead, use pmb.helpers.pmaports.get().

    :param must_exist: Raise an exception, when not found
    :param subpackages: set to False as speed optimization, if you know that
                        the package is not a subpackage of another package
                        (i.e. looking for UI packages for "pmbootstrap init").
                        If a previous search with subpackages=True has found
                        the package already, it will still be returned as
                        cached result.
    :returns: the full path to the package's dir in pmaports
    """
    # Try to get a cached result first (we assume that the aports don't change
    # in one pmbootstrap call)
    ret: Path | None = None
    # Sanity check
    if "*" in package:
        raise RuntimeError(f"Invalid pkgname: {package}")

    # Try to find an APKBUILD with the exact pkgname we are looking for
    path = _find_apkbuilds(with_extra_repos).get(package)
    if path:
        logging.verbose(f"{package}: found apkbuild: {path}")
        ret = path.parent
    elif subpackages:
        # No luck, take a guess what APKBUILD could have the package we are
        # looking for as subpackage
        guess = guess_main(package)
        # Parse the APKBUILD and verify if the guess was right
        if guess and _find_package_in_apkbuild(package, guess / "APKBUILD"):
            ret = guess

        if not ret:
            # Otherwise parse all APKBUILDs (takes time!), is the
            # package we are looking for a subpackage of any of those?
            for path_current in _find_apkbuilds().values():
                if _find_package_in_apkbuild(package, path_current):
                    ret = path_current.parent
                    break

        # If we still didn't find anything, as last resort: assume our
        # initial guess was right and the APKBUILD parser just didn't
        # find the subpackage in there because it is behind shell logic
        # that we don't parse.
        if not ret:
            ret = guess

    # Crash when necessary
    if ret is None and must_exist:
        show_pkg_not_found_systemd_hint(package, with_extra_repos)
        raise RuntimeError(f"Could not find package '{package}' in pmaports")

    return ret


def find_optional(package: str) -> Path | None:
    try:
        return find(package)
    except RuntimeError:
        return None


# The only caller with subpackages=False is ui.check_option()
@Cache("pkgname", "with_extra_repos", subpackages=True)
def get_with_path(
    pkgname: str,
    must_exist: bool = True,
    subpackages: bool = True,
    with_extra_repos: WithExtraRepos = "default",
) -> tuple[Path | None, Apkbuild | None]:
    """
    Find and parse an APKBUILD file.

    Run 'pmbootstrap apkbuild_parse hello-world' for a full output example.
    Relevant variables are defined in pmb.config.apkbuild_attributes.

    :param pkgname: the package name [+ version constraint] to find
    :param must_exist: raise an exception when it can't be found
    :param subpackages: also search for subpackages with the specified
        names (slow! might need to parse all APKBUILDs to find it)
    :param with_extra_repos: use extra repositories (e.g. systemd) when
        searching for the package

    :returns: relevant variables from the APKBUILD as dictionary, e.g.:
                  { "pkgname": "hello-world",
                  "arch": ["all"],
                  "pkgrel": "4",
                  "pkgrel": "1",
                  "options": [],
                  ... }
    """
    pkgname_no_op = pmb.helpers.package.remove_operators(pkgname)
    pmaport = find(pkgname_no_op, must_exist, subpackages, with_extra_repos)
    if pmaport:
        apkbuild = pmb.parse.apkbuild(pmaport / "APKBUILD")
        if pkgname_no_op == pkgname or pmb.helpers.package.check_version_constraints(
            pkgname, apkbuild["pkgver"]
        ):
            return pmaport, pmb.parse.apkbuild(pmaport / "APKBUILD")
    return None, None


@overload
def get(
    pkgname: str,
    must_exist: Literal[True] = ...,
    subpackages: bool = ...,
    with_extra_repos: WithExtraRepos = ...,
) -> Apkbuild: ...


@overload
def get(
    pkgname: str,
    must_exist: bool = ...,
    subpackages: bool = ...,
    with_extra_repos: WithExtraRepos = ...,
) -> Apkbuild | None: ...


def get(
    pkgname: str,
    must_exist: bool = True,
    subpackages: bool = True,
    with_extra_repos: WithExtraRepos = "default",
) -> Apkbuild | None:
    return get_with_path(pkgname, must_exist, subpackages, with_extra_repos)[1]


def find_providers(provide: str, default: list[str]) -> list[tuple[Any, Any]]:
    """
    Search for providers of the specified (virtual) package in pmaports.

    Note: Currently only providers from a single APKBUILD are returned.

    :param provide: the (virtual) package to search providers for
    :param default: the _pmb_default to look through for defaults
    :returns: tuple list (pkgname, apkbuild_pkg) with providers, sorted by
              provider_priority. The provider with the highest priority
              (which would be selected by default) comes first.
    """
    providers = {}

    apkbuild = get(provide)
    for subpkgname, subpkg in apkbuild["subpackages"].items():
        if subpkg is None:
            # subpackage might be for an abuild split func
            continue
        for provides in subpkg["provides"]:
            # Strip provides version (=$pkgver-r$pkgrel)
            if provides.split("=", 1)[0] == provide:
                if subpkgname in default:
                    subpkg["provider_priority"] = 999999
                providers[subpkgname] = subpkg

    return sorted(providers.items(), reverse=True, key=lambda p: p[1].get("provider_priority", 0))


def get_repo(pkgname: str) -> str | None:
    """
    Get the repository folder of an aport.

    :pkgname: package name
    :returns: * None if pkgname is not in extra-repos/
              * "systemd" if the pkgname is in extra-repos/systemd/
    """
    aport: Path
    aport = find(pkgname)

    if aport.parent.parent.name == "extra-repos":
        return aport.parent.name

    return None


def check_arches(arches: list[str], arch: Arch) -> bool:
    """
    Check if building for a certain arch is allowed.

    :param arches: list of all supported arches, as it can be found in the
        arch="" line of APKBUILDS (including all, noarch, !arch, ...).
        For example: ["x86_64", "x86", "!armhf"]

    :param arch: the architecture to check for

    :returns: True when building is allowed, False otherwise
    """
    if f"!{arch}" in arches:
        return False
    return any(value in arches for value in [str(arch), "all", "noarch"])


def get_channel_new(channel: str) -> str:
    """
    Translate legacy channel names to the new ones.
    Legacy names are still supported for compatibility with old branches (pmb#2015).

    :param channel: name as read from pmaports.cfg or channels.cfg, like "edge", "v21.03" etc., or potentially a legacy name like "stable".
    :returns: name in the new format, e.g. "edge" or "v21.03"
    """
    legacy_cfg = pmb.config.pmaports_channels_legacy
    if channel in legacy_cfg:
        ret = legacy_cfg[channel]
        logging.verbose(f"Legacy channel '{channel}' translated to '{ret}'")
        return ret
    return channel



# Copyright 2026 Oliver Smith, Paul Adam
# SPDX-License-Identifier: GPL-3.0-or-later

import pmb.build
from pmb.core import Chroot


def build_init(chroot: Chroot) -> None:
    pmb.build.init(chroot)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from pathlib import Path

import pmb.build
import pmb.chroot
import pmb.helpers.cli
import pmb.parse
from pmb.core import Chroot
from pmb.core.pkgrepo import pkgrepo_default_path
from pmb.helpers import logging
from pmb.types import PathString


def newapkbuild(folder: PathString, args_passed: list[str], force: bool = False) -> None:
    # Initialize build environment and build folder
    pmb.build.init()
    pmb.chroot.init(Chroot.native())
    build = Path("/home/pmos/build")
    build_outside = Chroot.native() / build
    if build_outside.exists():
        pmb.chroot.root(["rm", "-r", build])
    pmb.chroot.user(["mkdir", "-p", build])

    # Run newapkbuild
    pmb.chroot.user(["newapkbuild", *args_passed], working_dir=build)
    glob_result = list(build_outside.glob("*/APKBUILD"))
    if not len(glob_result):
        return

    # Paths for copying
    source_apkbuild = glob_result[0]
    pkgname = pmb.parse.apkbuild(source_apkbuild, False)["pkgname"]
    target = pkgrepo_default_path() / folder / pkgname

    # Move /home/pmos/build/$pkgname/* to /home/pmos/build/*
    for path in build_outside.glob("*/*"):
        path_inside = build / pkgname / os.path.basename(path)
        pmb.chroot.user(["mv", path_inside, build])
    pmb.chroot.user(["rmdir", build / pkgname])

    # Overwrite confirmation
    if os.path.exists(target):
        logging.warning(f"WARNING: Folder already exists: {target}")
        question = "Continue and delete its contents?"
        if not force and not pmb.helpers.cli.confirm(question):
            raise RuntimeError("Aborted.")
        pmb.helpers.run.user(["rm", "-r", target])

    # Copy the aport (without the extracted src folder)
    logging.info(f"Create {target}")
    pmb.helpers.run.user(["mkdir", "-p", target])
    for path in build_outside.glob("*"):
        if not path.is_dir():
            pmb.helpers.run.user(["cp", path, target])



# Copyright 2023 Clayton Craft
# SPDX-License-Identifier: GPL-3.0-or-later
import os

import pmb.helpers.package
import pmb.helpers.pmaports
import pmb.parse
from pmb.core.arch import Arch
from pmb.core.pkgrepo import pkgrepo_iglob
from pmb.helpers import logging
from pmb.types import WithExtraRepos


def list_ui(arch: Arch) -> list[tuple[str, str]]:
    """
    Get all UIs, for which aports are available with their description.

    :param arch: device architecture, for which the UIs must be available
    :returns: [("none", "No graphical..."), ("weston", "Wayland reference...")]
    """
    ret = [
        (
            "none",
            "Bare minimum OS image for testing and manual"
            ' customization. The "console" UI should be selected if'
            " a graphical UI is not desired.",
        )
    ]
    for path in sorted(pkgrepo_iglob("main/postmarketos-ui-*")):
        try:
            apkbuild = pmb.parse.apkbuild(path)
        except FileNotFoundError as exception:
            logging.debug("Skipping UI directory without APKBUILD '%s' (%s)", path, exception)
            continue
        ui = os.path.basename(path).split("-", 2)[2]
        if pmb.helpers.package.check_arch(apkbuild["pkgname"], arch):
            ret.append((ui, apkbuild["pkgdesc"]))
    return ret


def check_option(
    ui: str, option: str, must_exist: bool = True, with_extra_repos: WithExtraRepos = "default"
) -> bool:
    """
    Check if an option, such as pmb:systemd, is inside an UI's APKBUILD.

    If must_exist is set to False, False will be returned if the UI doesn't exist.
    """
    if ui == "none":
        # Users can select "none" as UI in "pmbootstrap init", which does not
        # have a UI package.
        return False

    pkgname = f"postmarketos-ui-{ui}"
    apkbuild = pmb.helpers.pmaports.get(
        pkgname, must_exist, subpackages=False, with_extra_repos=with_extra_repos
    )
    return option in apkbuild["options"] if apkbuild is not None else False



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
from getpass import getpass

import pmb.chroot.other
import pmb.config
import pmb.install
import pmb.parse
from pmb.core import Chroot, ChrootType
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.types import PmbArgs


def _parse_flavor(device: str, autoinstall: bool = True) -> str:
    """
    Verify the flavor argument if specified, or return a default value.

    :param autoinstall: make sure that at least one kernel flavor is installed
    """
    # Install a kernel and get its "flavor", where flavor is a pmOS-specific
    # identifier that is typically in the form
    # "postmarketos-<manufacturer>-<device/chip>", e.g.
    # "postmarketos-qcom-sdm845"
    chroot = Chroot(ChrootType.ROOTFS, device)
    flavor = pmb.chroot.other.kernel_flavor_installed(chroot, autoinstall)

    if not flavor:
        raise RuntimeError(
            f"No kernel flavors installed in chroot '{chroot}'! Please let"
            " your device package depend on a package starting with 'linux-'."
        )
    return flavor


def _parse_suffix(args: PmbArgs) -> Chroot:
    deviceinfo = pmb.parse.deviceinfo()
    if getattr(args, "image", None):
        rootfs = Chroot.native() / f"home/pmos/rootfs/{deviceinfo.codename}.img"
        return Chroot(ChrootType.IMAGE, str(rootfs))
    if getattr(args, "rootfs", None):
        return Chroot(ChrootType.ROOTFS, get_context().config.device)
    elif args.buildroot:
        if args.buildroot == "device":
            return Chroot.buildroot(deviceinfo.arch)
        else:
            return Chroot.buildroot(Arch.from_str(args.buildroot))
    elif args.suffix:
        (t_, s) = args.suffix.split("_")
        t: ChrootType = ChrootType(t_)
        return Chroot(t, s)
    else:
        return Chroot(ChrootType.NATIVE)


def install(args: PmbArgs) -> None:
    config = get_context().config
    device = config.device
    deviceinfo = pmb.parse.deviceinfo(device)
    is_split = args.split if args.split is not None else False
    if args.no_fde:
        logging.warning("WARNING: --no-fde is deprecated, as it is now the default.")
    if args.rsync and args.full_disk_encryption:
        raise ValueError("Installation using rsync is not compatible with full disk encryption.")
    if args.rsync and not args.disk:
        raise ValueError("Installation using rsync only works with --disk.")

    if args.rsync and args.filesystem == "btrfs":
        raise ValueError("Installation using rsync is not currently supported on btrfs filesystem.")

    if not args.disk and args.split is None:
        # Default to split if the flash method requires it
        flasher = pmb.config.flashers.get(deviceinfo.flash_method, {})
        if flasher.get("split", False):
            is_split = True

    # Android recovery zip related
    if args.android_recovery_zip and args.filesystem:
        raise ValueError(
            "--android-recovery-zip cannot be combined with --filesystem (patches welcome)"
        )
    if args.android_recovery_zip and args.full_disk_encryption:
        logging.info(
            "WARNING: --fde is rarely used in combination with"
            " --android-recovery-zip. If this does not work, consider"
            " using another method (e.g. installing via netcat)"
        )
        logging.info(
            "WARNING: the kernel of the recovery system (e.g. TWRP)"
            f" must support the cryptsetup cipher '{args.cipher}'."
        )
        logging.info(
            "If you know what you are doing, consider setting a"
            " different cipher with 'pmbootstrap install --cipher=..."
            " --fde --android-recovery-zip'."
        )

    # Don't install locally compiled packages and package signing keys
    if not args.install_local_pkgs:
        # Implies that we don't build outdated packages (overriding the answer
        # in 'pmbootstrap init')
        config.build_pkgs_on_install = False

        # Safest way to avoid installing local packages is having none
        if list((config.work / "packages").glob("*")):
            raise ValueError(
                "--no-local-pkgs specified, but locally built"
                " packages found. Consider 'pmbootstrap zap -p'"
                " to delete them."
            )

    if not args.password:
        args.password = getpass(f"Choose a password for the user '{config.user}': ")
        if getpass(f"Confirm password for '{config.user}': ") != args.password:
            raise NonBugError("Passwords did not match!")

    # Verify that the root filesystem is supported by current pmaports branch
    pmb.install.get_root_filesystem(args.filesystem)

    pmb.install.install(args, is_split)



# Copyright 2024 Stefan Hansson
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

from dataclasses import dataclass

import pmb.build._package
from pmb.core.apkindex_block import ApkindexBlock
from pmb.core.context import get_context
from pmb.types import Apkbuild


@dataclass
class PackageMetadata:
    # This can't be list[Arch] because it can have values like "noarch" and "!armhf"
    arch: list[str]
    depends: list[str]
    pkgname: str
    provides: list[str]
    version: str
    from_pmaports: bool

    @staticmethod
    def from_apkindex_block(apkindex_block: ApkindexBlock) -> PackageMetadata:
        return PackageMetadata(
            arch=[str(apkindex_block.arch)],
            depends=apkindex_block.depends,
            pkgname=apkindex_block.pkgname,
            provides=apkindex_block.provides,
            version=apkindex_block.version,
            from_pmaports=False,
        )

    @staticmethod
    def from_pmaport(pmaport: Apkbuild) -> PackageMetadata:
        pmaport_arches = pmaport["arch"]
        pmaport_depends = pmb.build._package.get_depends(get_context(), pmaport)
        pmaport_pkgname = pmaport["pkgname"]
        pmaport_provides = pmaport["provides"]
        pmaport_version = pmaport["pkgver"] + "-r" + pmaport["pkgrel"]

        return PackageMetadata(
            arch=pmaport_arches,
            depends=pmaport_depends or [],
            pkgname=pmaport_pkgname,
            provides=pmaport_provides,
            version=pmaport_version,
            from_pmaports=True,
        )



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from pathlib import Path

import pmb.config
import pmb.helpers.cli
import pmb.helpers.mount
import pmb.install.losetup
from pmb.core import Chroot
from pmb.core.context import get_context
from pmb.helpers import logging


def previous_install(path: Path) -> bool:
    """
    Search the disk for possible existence of a previous installation of
    pmOS. We temporarily mount the possible pmOS_boot partition as
    /dev/diskp1 inside the native chroot to check the label from there.

    :param path: path to disk block device (e.g. /dev/mmcblk0)
    """
    label = ""
    for blockdevice_outside in [path.with_stem(f"{path.name}1"), path.with_stem(f"{path.name}p1")]:
        if not blockdevice_outside.exists():
            continue
        blockdevice_inside = "/dev/diskp1"
        pmb.helpers.mount.bind_file(blockdevice_outside, Chroot.native() / blockdevice_inside)
        try:
            label = pmb.chroot.root(
                ["blkid", "-s", "LABEL", "-o", "value", blockdevice_inside], output_return=True
            )
        except RuntimeError:
            logging.info(
                "WARNING: Could not get block device label,"
                " assume no previous installation on that partition"
            )

        pmb.helpers.run.root(["umount", Chroot.native() / blockdevice_inside])
    return "pmOS_boot" in label


def mount_disk(path: Path) -> None:
    """:param path: path to disk block device (e.g. /dev/mmcblk0)"""
    # Sanity checks
    if not path.exists():
        raise RuntimeError(f"The disk block device does not exist: {path}")
    for path_mount in path.parent.glob(f"{path.name}*"):
        if pmb.helpers.mount.ismount(path_mount):
            raise RuntimeError(f"{path_mount} is mounted! Will not attempt to format this!")
    logging.info(f"(native) mount /dev/install (host: {path})")
    pmb.helpers.mount.bind_file(path, Chroot.native() / "dev/install")
    if previous_install(path):
        if not pmb.helpers.cli.confirm(
            "WARNING: This device has a previous installation of pmOS. CONTINUE?"
        ):
            raise RuntimeError("Aborted.")
    else:
        if not pmb.helpers.cli.confirm(f"EVERYTHING ON {path} WILL BE ERASED! CONTINUE?"):
            raise RuntimeError("Aborted.")


def create_and_mount_image(
    sector_size: int | None, size_boot: int, size_root: int, split: bool = False
) -> None:
    """
    Create a new image file, and mount it as /dev/install.

    :param size_boot: size of the boot partition in MiB
    :param size_root: size of the root partition in MiB
    :param split: create separate images for boot and root partitions
    """
    # Short variables for paths
    chroot = Chroot.native()
    config = get_context().config
    img_path_prefix = Path("/home/pmos/rootfs")
    img_path_full = img_path_prefix / f"{config.device}.img"
    img_path_boot = img_path_prefix / f"{config.device}-boot.img"
    img_path_root = img_path_prefix / f"{config.device}-root.img"

    # Umount and delete existing images
    for img_path in [img_path_full, img_path_boot, img_path_root]:
        outside = chroot / img_path
        if outside.exists():
            pmb.helpers.mount.umount_all(chroot / "mnt")
            pmb.install.losetup.umount(img_path)
            pmb.chroot.root(["rm", img_path])

    # Make sure there is enough free space
    size_mb = round(size_boot + size_root)
    disk_data = os.statvfs(get_context().config.work)
    free = round((disk_data.f_bsize * disk_data.f_bavail) / (1024**2))
    if size_mb > free:
        raise RuntimeError(
            f"Not enough free space to create rootfs image! (free: {free}M, required: {size_mb}M)"
        )

    # Create empty image files
    pmb.chroot.user(["mkdir", "-p", "/home/pmos/rootfs"])
    size_mb_full = str(size_mb) + "M"
    size_mb_boot = str(size_boot) + "M"
    size_mb_root = str(size_root) + "M"
    images = {img_path_full: size_mb_full}
    if split:
        images = {img_path_boot: size_mb_boot, img_path_root: size_mb_root}
    for img_path, image_size_mb in images.items():
        logging.info(f"(native) create {img_path.name} ({image_size_mb})")
        pmb.chroot.root(["truncate", "-s", image_size_mb, img_path])

    # Mount to /dev/install
    mount_image_paths = {img_path_full: "/dev/install"}
    if split:
        mount_image_paths = {img_path_boot: "/dev/installp1", img_path_root: "/dev/installp2"}

    for img_path, mount_point in mount_image_paths.items():
        logging.info(f"(native) mount {mount_point} ({img_path.name})")
        pmb.install.losetup.mount(img_path, sector_size)
        device = pmb.install.losetup.device_by_back_file(img_path)
        pmb.helpers.mount.bind_file(device, Chroot.native() / mount_point)


def create(sector_size: int | None, size_boot: int, size_root: int, split: bool) -> None:
    """
    Create /dev/install (the "install blockdevice").

    :param size_boot: size of the boot partition in MiB
    :param size_root: size of the root partition in MiB
    :param split: create separate images for boot and root partitions
    """
    pmb.helpers.mount.umount_all(Chroot.native() / "dev/install")
    create_and_mount_image(sector_size, size_boot, size_root, split)


def handle_disk_mount(disk: Path) -> None:
    """
    Mount disk as /dev/install (the "install blockdevice").

    :param disk: path to disk block device (e.g. /dev/mmcblk0).
    """
    pmb.helpers.mount.umount_all(Chroot.native() / "dev/install")
    mount_disk(disk)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import logging
import os
import sys
from pathlib import Path
from typing import Any, Final, TextIO

from pmb.meta import Cache

logfd: TextIO

CRITICAL: Final[int] = logging.CRITICAL
FATAL: Final[int] = logging.FATAL
ERROR: Final[int] = logging.ERROR
WARNING: Final[int] = logging.WARNING
WARN: Final[int] = logging.WARNING
INFO: Final[int] = logging.INFO
DEBUG: Final[int] = logging.DEBUG
NOTSET: Final[int] = logging.NOTSET
VERBOSE: Final[int] = 5


class LogHandler(logging.StreamHandler):
    """Write to stdout and to the already opened log file."""

    def __init__(self, details_to_stdout: bool = False, quiet: bool = False) -> None:
        super().__init__()
        self.details_to_stdout = details_to_stdout
        self.quiet = False

        # FIXME: importing pmb.config pulls in a whole lot of stuff
        # and can easily lead to circular imports, so we defer it until here.
        import pmb.config

        self.styles = pmb.config.styles

    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)

            # INFO or higher: Write to stdout
            if self.details_to_stdout or (not self.quiet and record.levelno >= logging.INFO):
                stream = self.stream

                styles = self.styles

                msg_col = (
                    msg.replace(
                        "NOTE:",
                        f"{styles['BLUE']}NOTE:{styles['END']}",
                        1,
                    )
                    .replace(
                        "WARNING:",
                        f"{styles['YELLOW']}WARNING:{styles['END']}",
                        1,
                    )
                    .replace(
                        "ERROR:",
                        f"{styles['RED']}ERROR:{styles['END']}",
                        1,
                    )
                    .replace(
                        "DONE!",
                        f"{styles['GREEN']}DONE!{styles['END']}",
                        1,
                    )
                    .replace(
                        "*** ",
                        f"{styles['GREEN']}*** ",
                        1,
                    )
                )

                for key, value in styles.items():
                    msg_col = msg_col.replace(f"@{key}@", value)
                    # Strip from the normal log message
                    msg = msg.replace(f"@{key}@", "")

                msg_col += styles["END"]

                stream.write(msg_col)
                stream.write(self.terminator)
                self.flush()

            # Everything: Write to logfd
            if not self.details_to_stdout:
                msg = "(" + str(os.getpid()).zfill(6) + ") " + msg
                logfd.write(msg + "\n")
                logfd.flush()

        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException:
            self.handleError(record)


def add_verbose_log_level() -> None:
    """
    Add a new log level "verbose", which is below "debug".

    Also monkeypatch logging, so it can be used with logging.verbose().

    This function is based on work by Voitek Zylinski and sleepycal:
    https://stackoverflow.com/a/20602183
    All stackoverflow user contributions are licensed as CC-BY-SA:
    https://creativecommons.org/licenses/by-sa/3.0/
    """
    logging.VERBOSE = VERBOSE  # type: ignore[attr-defined]
    logging.addLevelName(VERBOSE, "VERBOSE")
    logging.Logger.verbose = lambda inst, msg, *args, **kwargs: inst.log(  # type: ignore[attr-defined]
        VERBOSE, msg, *args, **kwargs
    )
    logging.verbose = lambda msg, *args, **kwargs: logging.log(VERBOSE, msg, *args, **kwargs)  # type: ignore[attr-defined]


def init(logfile: Path, verbose: bool, details_to_stdout: bool = False) -> None:
    """Set log format and add the log file descriptor to logfd, add the verbose log level."""
    global logfd

    if "logfs" in globals():
        warning("Logging already initialized, skipping...")
        return

    # Require containing directory to exist (so we don't create the work
    # folder and break the folder migration logic, which needs to set the
    # version upon creation)
    if not details_to_stdout and logfile.parent.exists():
        logfd = open(logfile, "a+")  # noqa: SIM115
        logfd.write("\n\n")
    elif details_to_stdout:
        logfd = sys.stdout
    else:
        logfd = open(os.devnull, "w")  # noqa: SIM115

    # Set log format
    root_logger = logging.getLogger()
    root_logger.handlers = []
    formatter = logging.Formatter("[%(asctime)s] %(message)s", datefmt="%H:%M:%S")

    # Set log level
    add_verbose_log_level()
    root_logger.setLevel(logging.DEBUG)
    if verbose:
        root_logger.setLevel(VERBOSE)

    # Add a custom log handler
    handler = LogHandler(details_to_stdout=details_to_stdout)
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)

    if "--password" in sys.argv:
        sys.argv[sys.argv.index("--password") + 1] = "[REDACTED]"
    logging.debug(f"$ pmbootstrap {' '.join(sys.argv)}")


def disable() -> None:
    logger = logging.getLogger()
    logger.disabled = True


# We have our own logging wrappers so we can make mypy happy
# by not calling the (undefined) logging.verbose() function.


def critical(msg: object, *args: object, **kwargs: Any) -> None:
    logging.critical(msg, *args, **kwargs)


def fatal(msg: object, *args: object, **kwargs: Any) -> None:
    logging.fatal(msg, *args, **kwargs)


def error(msg: object, *args: object, **kwargs: Any) -> None:
    logging.error(msg, *args, **kwargs)


def warning(msg: object, *args: object, **kwargs: Any) -> None:
    logging.warning(msg, *args, **kwargs)


@Cache("msg")
def warn_once(msg: str) -> None:
    logging.warning(msg)


def info(msg: object, *args: object, **kwargs: Any) -> None:
    logging.info(msg, *args, **kwargs)


def debug(msg: object, *args: object, **kwargs: Any) -> None:
    logging.debug(msg, *args, **kwargs)


def verbose(msg: object, *args: object, **kwargs: Any) -> None:
    logging.verbose(msg, *args, **kwargs)  # type: ignore[attr-defined]


def log(level: int, msg: object, *args: object, **kwargs: Any) -> None:
    logging.log(level, msg, *args, **kwargs)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from pathlib import Path
from typing import Literal, TextIO

import pmb.chroot
import pmb.chroot.apk
import pmb.helpers.run
from pmb.core import Chroot
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.types import Bootimg, PathString


def is_dtb(path: PathString) -> bool:
    if not os.path.isfile(path):
        return False
    with open(path, "rb") as f:
        # Check FDT magic identifier (0xd00dfeed)
        return f.read(4) == b"\xd0\x0d\xfe\xed"


def get_mtk_label(path: PathString) -> str | None:
    """
    Read the label from the MediaTek header of the kernel or ramdisk inside
    an extracted boot.img.

    :param path: to either the kernel or ramdisk extracted from boot.img
    :returns: * None: file does not exist or does not have MediaTek header

    * Label string (e.g. "ROOTFS", "KERNEL")
    """
    if not os.path.exists(path):
        return None

    with open(path, "rb") as f:
        # Check Mediatek header (0x88168858)
        if f.read(4) != b"\x88\x16\x88X":
            return None
        f.seek(8)
        label = f.read(32).decode("utf-8").rstrip("\0")

        if label == "RECOVERY":
            logging.warning(
                "WARNING: This boot.img has MediaTek headers. Since you passed a"
                " recovery image instead of a regular boot.img, we can't tell what"
                " the ramdisk signature label is supposed to be, so we assume that"
                " it's the most common value, ROOTFS. There is a chance that this"
                " is wrong and it may not boot; in that case, run bootimg_analyze"
                " again with a regular boot.img. If this *is* a regular boot.img,"
                " replace the value of deviceinfo_bootimg_mtk_label_ramdisk with"
                " 'RECOVERY'."
            )
            return "ROOTFS"
        else:
            return label


def get_qcdt_type(path: PathString) -> Literal["qcom", "sprd", "exynos"] | None:
    """
    Get the dt.img type by reading the first four bytes of the file.

    :param path: to the qcdt image extracted from boot.img
    :returns: * None: dt.img is of unknown type
    """
    if not os.path.exists(path):
        return None

    with open(path, "rb") as f:
        fourcc = f.read(4)

        match fourcc:
            case b"QCDT":
                return "qcom"
            case b"SPRD":
                return "sprd"
            case b"DTBH":
                return "exynos"
            case _:
                return None


def get_qcdt_exynos_platform_subtype(path: PathString) -> tuple[str, str] | tuple[None, None]:
    """
    Get the exynos dt.img platform and subtype by reading the first
    four bytes of the file.

    :param path: to the qcdt image extracted from boot.img
    :returns: ( Type tuple with platform and subtype, like ("0x50a6", "0x217584da")
    """
    if not os.path.exists(path):
        return (None, None)

    with open(path, "rb") as f:
        header = f.read(24)

    platform = hex(int.from_bytes(header[16:20], "little"))
    subtype = hex(int.from_bytes(header[20:24], "little"))

    return (platform, subtype)


def bootimg(path: Path) -> Bootimg:
    if not path.exists():
        raise RuntimeError(f"Could not find file '{path}'")

    logging.info(
        "NOTE: You will be prompted for your sudo/doas/run0 password, so"
        " we can set up a chroot to extract and analyze your"
        " boot.img file"
    )
    pmb.chroot.init(Chroot.native())
    pmb.chroot.apk.install(["file", "unpackbootimg"], Chroot.native())

    temp_path = Path("/tmp/bootimg_parser")
    pmb.chroot.user(["mkdir", "-p", temp_path])
    bootimg_path = Chroot.native() / temp_path / "boot.img"

    # Copy the boot.img into the chroot temporary folder
    # and make it world readable
    pmb.helpers.run.root(["cp", path, bootimg_path])
    pmb.helpers.run.root(["chmod", "a+r", bootimg_path])

    file_output = pmb.chroot.user(
        ["file", "-b", "boot.img"], working_dir=temp_path, output_return=True
    ).rstrip()
    if "android bootimg" not in file_output.lower():
        if get_context().force:
            logging.warning(
                "WARNING: boot.img file seems to be invalid, but proceeding anyway (-f specified)"
            )
        else:
            logging.info(
                "NOTE: If you are sure that your file is a valid"
                " boot.img file, you could force the analysis"
                f" with: 'pmbootstrap bootimg_analyze {path} -f'"
            )
            if (
                "linux kernel" in file_output.lower()
                or "ARM OpenFirmware FORTH dictionary" in file_output
            ):
                raise RuntimeError(
                    "File is a Kernel image, you might need the"
                    " 'heimdall-isorec' flash method. See also:"
                    " <https://wiki.postmarketos.org/wiki/"
                    "Deviceinfo_flash_methods>"
                )
            else:
                raise RuntimeError("File is not an Android boot.img. (" + file_output + ")")

    # Extract all the files
    pmb.chroot.user(["unpackbootimg", "-i", "boot.img"], working_dir=temp_path)

    output = {}
    header_version = 0
    # Get base, offsets, pagesize, cmdline and qcdt info
    # This file does not exist for example for qcdt images
    bootimg_header_version = Path(f"{bootimg_path}-header_version")
    if bootimg_header_version.exists():
        header_version = int(bootimg_header_version.read_text().replace("\n", ""))

    if header_version >= 3:
        output["pagesize"] = "4096"
    else:
        addresses = {
            "base": f"{bootimg_path}-base",
            "kernel_offset": f"{bootimg_path}-kernel_offset",
            "ramdisk_offset": f"{bootimg_path}-ramdisk_offset",
            "second_offset": f"{bootimg_path}-second_offset",
            "tags_offset": f"{bootimg_path}-tags_offset",
        }
        if header_version == 2:
            addresses["dtb_offset"] = f"{bootimg_path}-dtb_offset"
        for key, file in addresses.items():
            with open(file) as f:
                output[key] = f"0x{int(trim_input(f), 16):08x}"

        with open(f"{bootimg_path}-pagesize") as f:
            output["pagesize"] = trim_input(f)

    output["bootimg_qcdt"] = (
        "true"
        if os.path.isfile(f"{bootimg_path}-dt") and os.path.getsize(f"{bootimg_path}-dt") > 0
        else "false"
    )
    output.update(
        {
            key: value
            for key, value in {
                "mtk_label_kernel": get_mtk_label(f"{bootimg_path}-kernel"),
                "mtk_label_ramdisk": get_mtk_label(f"{bootimg_path}-ramdisk"),
                "bootimg_qcdt_type": get_qcdt_type(f"{bootimg_path}-dt"),
            }.items()
            if value is not None
        }
    )
    if "bootimg_qcdt_type" in output and output["bootimg_qcdt_type"] == "exynos":
        platform, subtype = get_qcdt_exynos_platform_subtype(f"{bootimg_path}-dt")
        # Omit if platform is default value 0x50a6
        if platform and platform != "0x50a6":
            output.update({"bootimg_qcdt_exynos_platform": platform})
        # Omit if subtype is default value 0x217584da
        if subtype and subtype != "0x217584da":
            output.update({"bootimg_qcdt_exynos_subtype": subtype})

    output["dtb_second"] = "true" if is_dtb(f"{bootimg_path}-second") else ""

    with open(f"{bootimg_path}-cmdline") as f:
        output["cmdline"] = trim_input(f)

    # Cleanup
    pmb.chroot.user(["rm", "-r", temp_path])

    return Bootimg(
        cmdline=output["cmdline"],
        bootimg_qcdt=output["bootimg_qcdt"],
        bootimg_qcdt_type=output.get("bootimg_qcdt_type"),
        bootimg_qcdt_exynos_platform=output.get("bootimg_qcdt_exynos_platform", ""),
        bootimg_qcdt_exynos_subtype=output.get("bootimg_qcdt_exynos_subtype", ""),
        dtb_offset=output.get("dtb_offset"),
        dtb_second=output["dtb_second"],
        base=output.get("base", ""),
        kernel_offset=output.get("kernel_offset", ""),
        ramdisk_offset=output.get("ramdisk_offset", ""),
        second_offset=output.get("second_offset", ""),
        tags_offset=output.get("tags_offset", ""),
        pagesize=output["pagesize"],
        header_version=header_version,
        mtk_label_kernel=output.get("mtk_label_kernel", ""),
        mtk_label_ramdisk=output.get("mtk_label_ramdisk", ""),
    )


def trim_input(f: TextIO) -> str:
    return f.read().replace("\n", "")



# Copyright 2026 Stefan Hansson, Martijn Braam
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import shlex
from pathlib import Path

import pmb.build
import pmb.config.pmaports
import pmb.helpers.run
import pmb.helpers.run_core
import pmb.parse.apkindex
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.types import PathString, RunOutputTypeDefault

su_cmd = "_su=$(command -v sudo >/dev/null && echo sudo || (command -v doas >/dev/null && echo doas || echo run0)); $_su"


def scp_abuild_key(user: str, host: str, port: str) -> None:
    """
    Copy the building key of the local installation to the target device,
    so it trusts the apks that were signed here.

    :param user: target device ssh username
    :param host: target device ssh hostname
    :param port: target device ssh port
    """
    keys = list((get_context().config.work / "config_abuild").glob("*.pub"))
    key = keys[0]
    key_name = os.path.basename(key)

    logging.info(f"Copying signing key ({key_name}) to {user}@{host}")
    command: list[PathString] = ["scp", "-P", port, key, f"{user}@{host}:/tmp"]
    pmb.helpers.run.user(command, output=RunOutputTypeDefault.INTERACTIVE)

    logging.info(f"Installing signing key at {user}@{host}")
    keyname = os.path.join("/tmp", os.path.basename(key))
    remote_cmd_l: list[PathString] = [
        "mv",
        "-n",
        keyname,
        "/etc/apk/keys/",
    ]
    remote_cmd = pmb.helpers.run_core.flat_cmd([remote_cmd_l])
    full_cmd = shlex.quote(f"{su_cmd} {remote_cmd}")
    command = ["ssh", "-t", "-p", port, f"{user}@{host}", f"sh -c {full_cmd}"]
    pmb.helpers.run.user(command, output=RunOutputTypeDefault.TUI)


def ssh_find_arch(user: str, host: str, port: str) -> Arch:
    """Connect to a device via ssh and query the architecture."""
    logging.info(f"Querying architecture of {user}@{host}")
    # Run command in a subshell in case the foreign device has a weird uname
    # implementation, e.g. Nushell.
    architecture_cmd = shlex.quote("uname -m")
    command = ["ssh", "-p", port, f"{user}@{host}", f"sh -c {architecture_cmd}"]
    output = pmb.helpers.run.user_output(command)
    # Split by newlines so we can pick out any irrelevant output, e.g. the "permanently
    # added to list of known hosts" warnings.
    output_lines = output.strip().splitlines()
    # Pick out last line which should contain the foreign device's architecture
    foreign_machine_type = output_lines[-1]
    return Arch.from_machine_type(foreign_machine_type)


def ssh_install_apks(
    user: str,
    host: str,
    port: str,
    paths: list[Path],
    install_offline: bool,
) -> None:
    """
    Copy binary packages via SCP and install them via SSH.

    :param user: target device ssh username
    :param host: target device ssh hostname
    :param port: target device ssh port
    :param paths: list of absolute paths to locally stored apks
    """
    remote_paths = [os.path.join("/tmp", os.path.basename(path)) for path in paths]

    logging.info(f"Copying packages to {user}@{host}")
    command: list[PathString] = ["scp", "-P", port, *paths, f"{user}@{host}:/tmp"]
    pmb.helpers.run.user(command, output=RunOutputTypeDefault.INTERACTIVE)

    logging.info(f"Installing packages at {user}@{host}")
    add_cmd_list = ["apk", "--wait", "30", "--timeout", "5"]
    if install_offline:
        add_cmd_list.append("--no-network")
    add_cmd_list.extend(["add", *remote_paths])
    add_cmd = pmb.helpers.run_core.flat_cmd([add_cmd_list])
    clean_cmd = pmb.helpers.run_core.flat_cmd([["rm", *remote_paths]])
    add_cmd_complete = shlex.quote(f"{su_cmd} {add_cmd} rc=$?; {clean_cmd} exit $rc")
    # Run apk command in a subshell in case the foreign device has a non-POSIX shell.
    command = ["ssh", "-t", "-p", port, f"{user}@{host}", f"sh -c {add_cmd_complete}"]
    pmb.helpers.run.user(command, output=RunOutputTypeDefault.TUI)


def sideload(
    user: str,
    host: str,
    port: str,
    arch: Arch | None,
    copy_key: bool,
    pkgnames: list[str],
) -> None:
    """
    Build packages if necessary and install them via SSH.

    :param user: target device ssh username
    :param host: target device ssh hostname
    :param port: target device ssh port
    :param arch: target device architecture
    :param copy_key: copy the abuild key too
    :param pkgnames: list of pkgnames to be built
    """
    paths = []

    if arch is None:
        arch = ssh_find_arch(user, host, port)

    context = get_context()
    to_build = []
    install_offline = True
    for pkgname in pkgnames:
        data_repo = pmb.parse.apkindex.package(pkgname, arch, True)

        if data_repo is None:
            raise RuntimeError(f"Couldn't find APKINDEX data for {pkgname}!")

        base_aports, _ = pmb.build.get_apkbuild(pkgname)
        channel = pmb.config.pmaports.read_config(base_aports)["channel"]

        apk_file = f"{pkgname}-{data_repo.version}.apk"
        host_path = context.config.work / "packages" / channel / arch / apk_file

        # If none of the packages have dependencies then we can run apk with
        # --no-network to avoid spurious APKINDEX updates
        if data_repo.depends:
            install_offline = False

        if not host_path.is_file():
            to_build.append(pkgname)

        paths.append(host_path)

    if to_build:
        pmb.build.packages(context, to_build, arch, force=True)
        # Check all the packages actually got builts
        for path in paths:
            if not path.is_file():
                raise RuntimeError(f"The package '{pkgname}' could not be built")

    if copy_key:
        scp_abuild_key(user, host, port)

    ssh_install_apks(user, host, port, paths, install_offline)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from pathlib import Path

import pmb.helpers.cli
import pmb.helpers.devices
import pmb.helpers.run
import pmb.parse
from pmb.core.arch import Arch
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.types import Bootimg


def ask_for_architecture() -> Arch:
    architectures = [str(a) for a in Arch.supported()]
    # Don't show armhf, new ports shouldn't use this architecture
    if "armhf" in architectures:
        architectures.remove("armhf")
    while True:
        ret = pmb.helpers.cli.ask(
            "Device architecture", architectures, "aarch64", complete=architectures
        )
        if ret in architectures:
            return Arch.from_str(ret)
        logging.fatal(
            "ERROR: Invalid architecture specified. If you want to add a new architecture, edit "
            "build_device_architectures in pmb/config/__init__.py."
        )


def ask_for_manufacturer() -> str:
    logging.info("Who produced the device (e.g. LG)?")
    return pmb.helpers.cli.ask("Manufacturer", None, None, False)


def ask_for_name(manufacturer: str) -> str:
    logging.info("What is the official name (e.g. Google Nexus 5)?")
    ret = pmb.helpers.cli.ask("Name", None, None, False)

    # Always add the manufacturer
    if not ret.startswith(manufacturer) and not ret.startswith("Google"):
        ret = manufacturer + " " + ret
    return ret


def ask_for_year() -> str:
    # Regex from https://stackoverflow.com/a/12240826
    logging.info("In what year was the device released (e.g. 2012)?")
    return pmb.helpers.cli.ask("Year", None, None, False, validation_regex=r"^[1-9]\d{3,}$")


def ask_for_chassis() -> str:
    types = pmb.config.deviceinfo_chassis_types

    logging.info("What type of device is it?")
    logging.info("Valid types are: " + ", ".join(types))
    logging.info("Leave empty if this property is defined elsewhere, e.g. in Devicetree or ACPI")
    return pmb.helpers.cli.ask(
        "Chassis", None, None, True, validation_regex="|".join([*types, "None"]), complete=types
    )


def ask_for_flash_method() -> str:
    while True:
        logging.info("Which flash method does the device support?")
        method = pmb.helpers.cli.ask(
            "Flash method", pmb.config.flash_methods, "none", complete=pmb.config.flash_methods
        )

        if method in pmb.config.flash_methods:
            if method == "heimdall":
                heimdall_types = ["isorec", "bootimg"]
                while True:
                    logging.info('Does the device use the "isolated recovery" or boot.img?')
                    logging.info(
                        "<https://wiki.postmarketos.org/wiki/Deviceinfo_flash_methods#Isorec_or_bootimg.3F>"
                    )
                    heimdall_type = pmb.helpers.cli.ask("Type", heimdall_types, heimdall_types[0])
                    if heimdall_type in heimdall_types:
                        method += "-" + heimdall_type
                        break
                    logging.fatal("ERROR: Invalid type specified.")
            return method

        logging.fatal(
            "ERROR: Invalid flash method specified. If you want to add a new flash method, edit "
            "flash_methods in pmb/config/__init__.py."
        )


def ask_for_bootimg() -> Bootimg | None:
    logging.info(
        "You can analyze a known working boot.img file to automatically fill out the flasher "
        "information for your deviceinfo file. Either specify the path to an image or press return "
        "to skip this step (you can do it later with 'pmbootstrap bootimg_analyze')."
    )

    while True:
        response = pmb.helpers.cli.ask("Path", None, "", False)
        if not response:
            return None
        path = Path(os.path.expanduser(response))
        try:
            return pmb.parse.bootimg(path)
        except Exception as e:
            logging.fatal("ERROR: " + str(e) + ". Please try again.")


def generate_deviceinfo_fastboot_content(bootimg: Bootimg | None = None) -> str:
    if bootimg is None:
        bootimg = Bootimg(
            cmdline="",
            bootimg_qcdt="false",
            bootimg_qcdt_type=None,
            bootimg_qcdt_exynos_platform=None,
            bootimg_qcdt_exynos_subtype=None,
            dtb_offset=None,
            dtb_second="false",
            base="",
            kernel_offset="",
            ramdisk_offset="",
            second_offset="",
            tags_offset="",
            pagesize="2048",
            header_version=None,
            mtk_label_kernel="",
            mtk_label_ramdisk="",
        )

    content = f"""\
        deviceinfo_kernel_cmdline="{bootimg["cmdline"]}"
        deviceinfo_generate_bootimg="true"
        deviceinfo_flash_pagesize="{bootimg["pagesize"]}"
        """

    for k in [
        "bootimg_qcdt",
        "bootimg_qcdt_type",
        "bootimg_qcdt_exynos_platform",
        "bootimg_qcdt_exynos_subtype",
        "dtb_second",
        "mtk_label_kernel",
        "mtk_label_ramdisk",
        "header_version",
    ]:
        v = bootimg[k]  # type: ignore[literal-required]
        if v:
            content += f"""\
            deviceinfo_{k}="{v}"
            """

    if bootimg["header_version"] == 2:
        content += f"""\
        deviceinfo_append_dtb="false"
        deviceinfo_flash_offset_dtb="{bootimg["dtb_offset"]}"
        """

    if bootimg["base"]:
        content += f"""\
        deviceinfo_flash_offset_base="{bootimg["base"]}"
        deviceinfo_flash_offset_kernel="{bootimg["kernel_offset"]}"
        deviceinfo_flash_offset_ramdisk="{bootimg["ramdisk_offset"]}"
        deviceinfo_flash_offset_second="{bootimg["second_offset"]}"
        deviceinfo_flash_offset_tags="{bootimg["tags_offset"]}"
        """

    return content


def generate_deviceinfo(
    pkgname: str,
    name: str,
    manufacturer: str,
    year: str,
    arch: Arch,
    chassis: str,
    flash_method: str,
    bootimg: Bootimg | None = None,
) -> None:
    codename = "-".join(pkgname.split("-")[1:])
    # Note: New variables must be added to pmb/config/__init__.py as well
    content = f"""\
        # Reference: <https://postmarketos.org/deviceinfo>
        # Please use double quotes only. You can source this file in shell
        # scripts.

        deviceinfo_format_version="0"
        deviceinfo_name="{name}"
        deviceinfo_manufacturer="{manufacturer}"
        deviceinfo_codename="{codename}"
        deviceinfo_year="{year}"
        deviceinfo_dtb=""
        deviceinfo_arch="{arch}"

        # Device related
        """

    content += f'deviceinfo_chassis="{chassis}"\n' if chassis != "None" else ""

    content += f"""\

        # Bootloader related
        deviceinfo_flash_method="{flash_method}"
        """

    content_heimdall_bootimg = """\
        deviceinfo_flash_heimdall_partition_kernel=""
        deviceinfo_flash_heimdall_partition_rootfs=""
        """

    content_heimdall_isorec = """\
        deviceinfo_flash_heimdall_partition_kernel=""
        deviceinfo_flash_heimdall_partition_initfs=""
        deviceinfo_flash_heimdall_partition_rootfs=""
        """

    content_0xffff = """\
        deviceinfo_generate_legacy_uboot_initfs="true"
        """

    content_uuu = """\
        deviceinfo_generate_legacy_uboot_initfs="true"
        """

    if flash_method == "fastboot":
        content += generate_deviceinfo_fastboot_content(bootimg)
    elif flash_method == "heimdall-bootimg":
        content += generate_deviceinfo_fastboot_content(bootimg)
        content += content_heimdall_bootimg
    elif flash_method == "heimdall-isorec":
        content += content_heimdall_isorec
    elif flash_method == "0xffff":
        content += content_0xffff
    elif flash_method == "uuu":
        content += content_uuu

    # Write to file
    work = get_context().config.work
    pmb.helpers.run.user(["mkdir", "-p", work / "aportgen"])
    path = work / "aportgen/deviceinfo"
    with open(path, "w", encoding="utf-8") as handle:
        handle.writelines(line.lstrip() + "\n" for line in content.rstrip().split("\n"))


def generate_modules_initfs() -> None:
    content = """\
    # Remove this file if unnecessary (CHANGEME!)
    # This file shall contain a list of modules to be included in the initramfs,
    # so that they are available in early boot stages. In general, it should
    # include modules to support unlocking FDE (touchscreen, panel, etc),
    # USB networking, and telnet in the debug-shell.
    # The format is one module name per line. Lines starting with the character
    # '#', and empty lines are ignored. If there are multiple kernel variants
    # with different initramfs module requirements, one modules-initfs.$variant
    # file should be created for each of them.
    """

    # Write to file
    work = get_context().config.work
    pmb.helpers.run.user(["mkdir", "-p", work / "aportgen"])
    path = work / "aportgen/modules-initfs"
    with open(path, "w", encoding="utf-8") as handle:
        handle.writelines(line.lstrip() + "\n" for line in content.rstrip().split("\n"))


def generate_apkbuild(
    pkgname: str,
    name: str,
    arch: Arch,
    flash_method: str,
    device_category: pmb.helpers.devices.DeviceCategory,
) -> None:
    # Dependencies
    depends = ["postmarketos-base"]
    if device_category == pmb.helpers.devices.DeviceCategory.DOWNSTREAM:
        depends += ["linux-" + "-".join(pkgname.split("-")[1:]), "postmarketos-base-downstream"]
    else:
        depends += ["linux-CHANGEME"]

    if flash_method in ["fastboot", "heimdall-bootimg"]:
        depends.append("mkbootimg")
    if flash_method == "0xffff":
        depends.append("uboot-tools")

    # Whole APKBUILD
    depends.sort()
    depends_fmt = ("\n" + " " * 12).join(depends)
    content = f"""\
        # Reference: <https://postmarketos.org/devicepkg>
        maintainer=""
        pkgname={pkgname}
        pkgdesc="{name}"
        pkgver=1
        pkgrel=0
        url="https://postmarketos.org"
        license="MIT"
        arch="{arch}"
        options="!check !archcheck"
        depends="
            {depends_fmt}
        "
        makedepends="devicepkg-dev"
        source="
            deviceinfo
            modules-initfs
        "

        build() {{
            devicepkg_build $startdir $pkgname
        }}

        package() {{
            devicepkg_package $startdir $pkgname
        }}

        sha512sums="(run 'pmbootstrap checksum {pkgname}' to fill)"
        """

    # Write the file
    work = get_context().config.work
    pmb.helpers.run.user(["mkdir", "-p", work / "aportgen"])
    path = work / "aportgen/APKBUILD"
    with open(path, "w", encoding="utf-8") as handle:
        handle.writelines(
            line[8:].replace(" " * 4, "\t") + "\n" for line in content.rstrip().split("\n")
        )


def generate(pkgname: str, device_category: pmb.helpers.devices.DeviceCategory) -> None:
    arch = ask_for_architecture()
    manufacturer = ask_for_manufacturer()
    name = ask_for_name(manufacturer)
    year = ask_for_year()
    chassis = ask_for_chassis()
    flash_method = ask_for_flash_method()
    bootimg = None
    if flash_method in ["fastboot", "heimdall-bootimg"]:
        bootimg = ask_for_bootimg()

    generate_deviceinfo(
        pkgname,
        name,
        manufacturer,
        year,
        arch,
        chassis,
        flash_method,
        bootimg,
    )
    generate_modules_initfs()
    generate_apkbuild(pkgname, name, arch, flash_method, device_category)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.aportgen.core
import pmb.build
import pmb.helpers.repo
import pmb.parse.apkindex
from pmb.core import Chroot
from pmb.core.arch import Arch


def generate(pkgname: str) -> None:
    arch = Arch.from_str(pkgname.split("-")[1])

    # Update or create APKINDEX for relevant arch so we know it exists and is recent.
    pmb.helpers.repo.update(arch)
    # Parse musl version from APKINDEX
    package_data = pmb.parse.apkindex.package("musl", arch=arch)
    if package_data is None:
        raise RuntimeError("Couldn't find package musl!")
    version = package_data.version
    pkgver = version.split("-r")[0]
    pkgrel = version.split("-r")[1]

    tempdir = pmb.aportgen.core.prepare_tempdir()

    # Write the APKBUILD
    channel_cfg = pmb.config.pmaports.read_config_channel()
    mirrordir = channel_cfg["mirrordir_alpine"]
    apkbuild_path = Chroot.native() / tempdir / "APKBUILD"
    apk_name = f"$srcdir/musl-$pkgver-r$pkgrel-$_arch-{mirrordir}.apk"
    apk_dev_name = f"$srcdir/musl-dev-$pkgver-r$pkgrel-$_arch-{mirrordir}.apk"
    with open(apkbuild_path, "w", encoding="utf-8") as handle:
        apkbuild = f"""\
            # Automatically generated aport, do not edit!
            # Generator: pmbootstrap aportgen {pkgname}

            # Stub for apkbuild-lint
            if [ -z "$(type -t arch_to_hostspec)" ]; then
                arch_to_hostspec() {{ :; }}
            fi

            maintainer=""
            pkgname={pkgname}
            pkgver={pkgver}
            pkgrel={pkgrel}
            arch="{pmb.aportgen.get_cross_package_arches(pkgname)}"
            subpackages="musl-dev-{arch}:package_dev"

            _arch="{arch}"
            _mirror="{pmb.config.aportgen_mirror_alpine}"

            url="https://musl-libc.org"
            license="MIT"
            options="!check !strip"
            pkgdesc="the musl library (lib c) implementation for $_arch"

            _target="$(arch_to_hostspec $_arch)"

            source="
                musl-$pkgver-r$pkgrel-$_arch-{mirrordir}.apk::$_mirror/{mirrordir}/main/$_arch/musl-$pkgver-r$pkgrel.apk
                musl-dev-$pkgver-r$pkgrel-$_arch-{mirrordir}.apk::$_mirror/{mirrordir}/main/$_arch/musl-dev-$pkgver-r$pkgrel.apk
            "

            package() {{
                mkdir -p "$pkgdir/usr/$_target"
                cd "$pkgdir/usr/$_target"
                # Use 'busybox tar' to avoid 'tar: Child returned status 141'
                # on some machines (builds.sr.ht, gitlab-ci). See pmaports#26.
                busybox tar -xf {apk_name}
                rm .PKGINFO .SIGN.*
            }}
            package_dev() {{
                mkdir -p "$subpkgdir/usr/$_target"
                cd "$subpkgdir/usr/$_target"
                # Use 'busybox tar' to avoid 'tar: Child returned status 141'
                # on some machines (builds.sr.ht, gitlab-ci). See pmaports#26.
                busybox tar -xf {apk_dev_name}
                rm .PKGINFO .SIGN.*

                # symlink everything from /usr/$_target/usr/*
                # to /usr/$_target/* so the cross-compiler gcc does not fail
                # to build.
                for _dir in include lib; do
                    mkdir -p "$subpkgdir/usr/$_target/$_dir"
                    cd "$subpkgdir/usr/$_target/usr/$_dir"
                    for i in *; do
                        cd "$subpkgdir/usr/$_target/$_dir"
                        ln -s /usr/$_target/usr/$_dir/$i $i
                    done
                done
            }}
        """
        handle.writelines(line[12:].replace(" " * 4, "\t") + "\n" for line in apkbuild.split("\n"))

    pmb.aportgen.core.generate_checksums(tempdir, apkbuild_path)



# Copyright 2024 Caleb Connolly
# SPDX-License-Identifier: GPL-3.0-or-later

import enum
from pathlib import Path

import pmb.chroot
import pmb.helpers.run
from pmb.core import Context
from pmb.core.arch import Arch
from pmb.core.chroot import Chroot
from pmb.core.pkgrepo import pkgrepo_name, pkgrepo_paths
from pmb.helpers import logging
from pmb.types import Apkbuild, CrossCompile, Env, RunOutputTypeDefault


class BootstrapStage(enum.IntEnum):
    """
    Pass a BOOTSTRAP= environment variable with the given value to abuild. See
    bootstrap_1 etc. at https://postmarketos.org/pmaports.cfg for details.
    """

    NONE = 0
    # We don't need explicit representations of the other numbers.


def override_source(
    apkbuild: Apkbuild, pkgver: str, src: str | None, chroot: Chroot = Chroot.native()
) -> None:
    """
    Mount local source inside chroot and append new functions (prepare() etc.)
    to the APKBUILD to make it use the local source.
    """
    if not src:
        return

    # Mount source in chroot
    mount_path = "/mnt/pmbootstrap/source-override/"
    mount_path_outside = chroot / mount_path
    pmb.helpers.mount.bind(src, mount_path_outside, umount=True)

    # Delete existing append file
    append_path = "/tmp/APKBUILD.append"
    append_path_outside = chroot / append_path
    if append_path_outside.exists():
        pmb.chroot.root(["rm", append_path], chroot)

    # Add src path to pkgdesc, cut it off after max length
    pkgdesc = ("[" + src + "] " + apkbuild["pkgdesc"])[:127]

    pkgname = apkbuild["pkgname"]

    # Appended content
    append = (
        """
             # ** Overrides below appended by pmbootstrap for --src **

             pkgver=\""""
        + pkgver
        + """\"
             pkgdesc=\""""
        + pkgdesc
        + """\"
             _pmb_src_copy="/tmp/pmbootstrap-local-source-copy"

             # Empty $source avoids patching in prepare()
             _pmb_source_original="$source"
             source=""
             sha512sums=""

             fetch() {
                 # Update source copy
                 msg "Copying source from host system: """
        + src
        + """\"
                 local exclude_from=\""""
        + mount_path
        + """/.gitignore\"
                 local rsync_args=""
                 if [ -f "$exclude_from" ]; then
                     rsync_args="--exclude-from=\"$exclude_from\""
                 fi
                 if ! [ \""""
        + pkgname
        + """\" = "$(cat /tmp/src-pkgname)" ]; then
                     rsync_args="--delete $rsync_args"
                 fi
                 rsync -a --exclude=".git/" $rsync_args --ignore-errors --force \\
                     \""""
        + mount_path
        + """\" "$_pmb_src_copy" || true

                 # Link local source files (e.g. kernel config)
                 mkdir "$srcdir"
                 local s
                 for s in $_pmb_source_original; do
                     is_remote "$s" || ln -sf "$startdir/$s" "$srcdir/"
                 done

                 echo \""""
        + pkgname
        + """\" > /tmp/src-pkgname
             }

             unpack() {
                 ln -sv "$_pmb_src_copy" "$builddir"
             }
             """
    )

    # Write and log append file
    with open(append_path_outside, "w", encoding="utf-8") as handle:
        handle.writelines(line[13:].replace(" " * 4, "\t") + "\n" for line in append.split("\n"))
    pmb.chroot.user(["cat", append_path], chroot)

    # Append it to the APKBUILD
    apkbuild_path = "/home/pmos/build/APKBUILD"
    shell_cmd = "cat " + apkbuild_path + " " + append_path + " > " + append_path + "_"
    pmb.chroot.user(["sh", "-c", shell_cmd], chroot)
    pmb.chroot.user(["mv", append_path + "_", apkbuild_path], chroot)


def mount_pmaports(chroot: Chroot = Chroot.native()) -> dict[str, Path]:
    """
    Mount pmaports.git in chroot.

    :param chroot: chroot to target
    :returns: dictionary mapping pkgrepo name to dest path
    """
    dest_paths = {}
    for repo in pkgrepo_paths(skip_extras=True):
        destination = Path("/mnt") / pkgrepo_name(repo)
        outside_destination = chroot / destination
        pmb.helpers.mount.bind(repo, outside_destination, umount=True)
        dest_paths[pkgrepo_name(repo)] = destination

    return dest_paths


def link_to_git_dir(chroot: Chroot) -> None:
    """
    Make ``/home/pmos/build/.git`` point to the .git dir from pmaports.git, with a
    symlink so abuild does not fail (#1841).

    abuild expects the current working directory to be a subdirectory of a
    cloned git repository (e.g. main/openrc from aports.git). If git is
    installed, it will try to get the last git commit from that repository, and
    place it in the resulting apk (.PKGINFO) as well as use the date from that
    commit as SOURCE_DATE_EPOCH (for reproducible builds).

    With that symlink, we actually make it use the last git commit from
    pmaports.git for SOURCE_DATE_EPOCH and have that in the resulting apk's
    .PKGINFO.
    """
    # Mount pmaports.git in chroot, in case the user did not use pmbootstrap to
    # clone it (e.g. how we build on sourcehut). Do this here and not at the
    # initialization of the chroot, because the pmaports dir may not exist yet
    # at that point. Use umount=True, so we don't have an old path mounted
    # (some tests change the pmaports dir).
    dest_paths = mount_pmaports(chroot)

    # Create .git symlink
    pmb.chroot.user(["mkdir", "-p", "/home/pmos/build"], chroot)
    pmb.chroot.user(["ln", "-sf", dest_paths["pmaports"] / ".git", "/home/pmos/build/.git"], chroot)


def handle_csum_failure(apkbuild: Apkbuild, chroot: Chroot) -> None:
    csum_fail_path = chroot / "tmp/apkbuild_verify_failed"
    if not csum_fail_path.exists():
        return

    reason = csum_fail_path.open().read().strip()
    if reason == "local":
        logging.info(
            "WARNING: Some checksums didn't match, run"
            f" 'pmbootstrap checksum {apkbuild['pkgname']}' to fix them."
        )
    else:
        logging.error(f"ERROR: Remote checksum mismatch for {apkbuild['pkgname']}")
        logging.error("NOTE: If you just modified this package:")
        logging.error(
            f" * run 'pmbootstrap checksum {apkbuild['pkgname']}' to update the checksums."
        )
        logging.error("If you didn't modify it, try building again to re-download the sources.")
        raise RuntimeError(f"Remote checksum mismatch for {apkbuild['pkgname']}")


def run_abuild(
    context: Context,
    apkbuild: Apkbuild,
    pkgver: str,
    channel: str,
    arch: Arch,
    cross: CrossCompile,
    strict: bool = False,
    force: bool = False,
    src: str | None = None,
    bootstrap_stage: int = BootstrapStage.NONE,
) -> None:
    """
    Set up all environment variables and construct the abuild command (all
    depending on the cross-compiler method and target architecture), copy
    the aport to the chroot and execute abuild.

    :param src: override source used to build the package with a local folder
    :param bootstrap_stage: pass a BOOTSTRAP= env var with the value to abuild
    :returns: (output, cmd, env), output is the destination apk path relative
              to the package folder ("x86_64/hello-1-r2.apk"). cmd and env are
              used by the test case, and they are the full abuild command and
              the environment variables dict generated in this function.
    """
    # Sanity check
    if cross == CrossCompile.CROSS_NATIVE and "!tracedeps" not in apkbuild["options"]:
        logging.warning(
            "WARNING: Option !tracedeps is not set, but cross compiling with"
            " cross-native (version 1). This will probably fail!"
        )

    hostchroot = cross.host_chroot(arch)
    buildchroot = cross.build_chroot(arch)

    # For cross-native2 compilation, bindmount the "host" rootfs to /mnt/sysroot
    # it will be used as the "sysroot"
    if cross == CrossCompile.CROSS_NATIVE2:
        if buildchroot != Chroot.native():
            raise ValueError(
                "Trying to use cross-native2 build buildchroot != native! This is a bug"
            )
        pmb.mount.bind(hostchroot.path, buildchroot / "/mnt/sysroot", umount=True)

    pkgdir = context.config.work / "packages" / channel
    if not pkgdir.exists():
        pmb.helpers.run.root(["mkdir", "-p", pkgdir])
        pmb.helpers.run.root(
            [
                "chown",
                "-R",
                f"{pmb.config.chroot_uid_user}:{pmb.config.chroot_uid_user}",
                pkgdir.parent,
            ]
        )

    pmb.chroot.rootm(
        [
            ["mkdir", "-p", "/home/pmos/packages"],
            ["rm", "-f", "/home/pmos/packages/pmos"],
            ["ln", "-sf", f"/mnt/pmbootstrap/packages/{channel}", "/home/pmos/packages/pmos"],
        ],
        buildchroot,
    )

    # Environment variables
    env: Env = {"SUDO_APK": "abuild-apk --no-progress"}
    if cross == CrossCompile.CROSS_NATIVE:
        hostspec = arch.alpine_triple()
        env["CROSS_COMPILE"] = hostspec + "-"
        env["CC"] = hostspec + "-gcc"
    if cross == CrossCompile.CROSS_NATIVE2:
        env["CHOST"] = str(arch)
        env["CBUILDROOT"] = "/mnt/sysroot"
        env["CFLAGS"] = "-Wl,-rpath-link=/mnt/sysroot/usr/lib"
        env["CGO_CFLAGS"] = "--sysroot=/mnt/sysroot"
        env["CGO_LDFLAGS"] = "--sysroot=/mnt/sysroot"
        try:
            env["GOARCH"] = arch.go()
        except ValueError:
            logging.debug(f"Not setting $GOARCH for {arch}")

    elif cross == CrossCompile.CROSSDIRECT:
        env["PATH"] = ":".join([f"/native/usr/lib/crossdirect/{arch}", pmb.config.chroot_path])
    else:
        env["CARCH"] = str(arch)
    if not context.ccache:
        env["CCACHE_DISABLE"] = "1"

    # Use sccache without crossdirect (crossdirect uses it via rustc.sh)
    if context.ccache and cross != CrossCompile.CROSSDIRECT:
        env["RUSTC_WRAPPER"] = "/usr/bin/sccache"

    # Cache binary objects from go in this path (like ccache)
    env["GOCACHE"] = "/home/pmos/.cache/go-build"

    # Cache go modules (git repositories). Usually these should be bundled and
    # it should not be required to download them at build time, in that case
    # the APKBUILD sets the GOPATH (and therefore indirectly GOMODCACHE). But
    # e.g. when using --src they are not bundled, in that case it makes sense
    # to point GOMODCACHE at pmbootstrap's work dir so the modules are only
    # downloaded once.
    if context.go_mod_cache:
        env["GOMODCACHE"] = "/home/pmos/go/pkg/mod"

    if bootstrap_stage:
        env["BOOTSTRAP"] = str(bootstrap_stage)

    # Build the abuild command
    # Since we install dependencies with pmb, disable dependency handling in abuild.
    # This is also required so that abuild doesn't try to install base-build-$ARCH packages
    # which don't exist
    cmd = ["abuild", "-d", "-D", "postmarketOS"]
    if force:
        cmd += ["-f"]
    if src:
        # Keep build artifacts, so repeated invocations will do incremental
        # building.
        cmd += ["-K"]

    # Copy the aport to the chroot and build it
    pmb.build.copy_to_buildpath(apkbuild["pkgname"], buildchroot, no_override=strict)
    if src and strict:
        logging.debug(f"({buildchroot}) Ensuring previous build artifacts are removed")
        pmb.chroot.root(["rm", "-rf", "/tmp/pmbootstrap-local-source-copy"], buildchroot)
    override_source(apkbuild, pkgver, src, buildchroot)
    link_to_git_dir(buildchroot)

    try:
        pmb.chroot.user(cmd, buildchroot, Path("/home/pmos/build"), env=env)
    finally:
        handle_csum_failure(apkbuild, buildchroot)

    pmb.helpers.run.root(
        ["umount", buildchroot / "/mnt/sysroot"], output=RunOutputTypeDefault.NULL, check=False
    )



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
import re
from collections import OrderedDict
from pathlib import Path
from typing import Any

import pmb.config
import pmb.helpers.devices
import pmb.parse.version
from pmb.core.context import get_context
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError
from pmb.meta import Cache
from pmb.types import Apkbuild

# sh variable name regex: https://stackoverflow.com/a/2821201/3527128

# ${foo}
revar = re.compile(r"\${([a-zA-Z_]+[a-zA-Z0-9_]*)}")

# $foo
revar2 = re.compile(r"\$([a-zA-Z_]+[a-zA-Z0-9_]*)")

# ${var/foo/bar}, ${var/foo/}, ${var/foo} -- replace foo with bar
revar3 = re.compile(r"\${([a-zA-Z_]+[a-zA-Z0-9_]*)/([^/]+)(?:/([^/]*?))?}")

# ${foo#bar} -- cut off bar from foo from start of string
revar4 = re.compile(r"\${([a-zA-Z_]+[a-zA-Z0-9_]*)#(.*)}")

# foo=
revar5 = re.compile(r"([a-zA-Z_]+[a-zA-Z0-9_]*)=")


def replace_variable(apkbuild: Apkbuild, value: str) -> str:
    def log_key_not_found(match: re.Match) -> None:
        logging.verbose(
            f"{apkbuild['pkgname']}: key '{match.group(1)}' for"
            f" replacing '{match.group(0)}' not found, ignoring"
        )

    # ${foo}
    for match in revar.finditer(value):
        try:
            logging.verbose(
                "{}: replace '{}' with '{}'".format(
                    apkbuild["pkgname"], match.group(0), apkbuild[match.group(1)]
                )
            )
            value = value.replace(match.group(0), apkbuild[match.group(1)], 1)
        except KeyError:
            log_key_not_found(match)

    # $foo
    for match in revar2.finditer(value):
        try:
            newvalue = apkbuild[match.group(1)]
            logging.verbose(
                "{}: replace '{}' with '{}'".format(apkbuild["pkgname"], match.group(0), newvalue)
            )
            value = value.replace(match.group(0), newvalue, 1)
        except KeyError:
            log_key_not_found(match)

    # ${var/foo/bar}, ${var/foo/}, ${var/foo}
    for match in revar3.finditer(value):
        try:
            newvalue = apkbuild[match.group(1)]
            search = match.group(2)
            replacement = match.group(3)
            if replacement is None:  # arg 3 is optional
                replacement = ""
            newvalue = newvalue.replace(search, replacement, 1)
            logging.verbose(
                "{}: replace '{}' with '{}'".format(apkbuild["pkgname"], match.group(0), newvalue)
            )
            value = value.replace(match.group(0), newvalue, 1)
        except KeyError:
            log_key_not_found(match)

    # ${foo#bar}
    rematch4 = revar4.finditer(value)
    for match in rematch4:
        try:
            newvalue = apkbuild[match.group(1)]
            substr = match.group(2)
            if newvalue.startswith(substr):
                newvalue = newvalue.replace(substr, "", 1)
            logging.verbose(
                "{}: replace '{}' with '{}'".format(apkbuild["pkgname"], match.group(0), newvalue)
            )
            value = value.replace(match.group(0), newvalue, 1)
        except KeyError:
            log_key_not_found(match)

    return value


def function_body(path: Path, func: str) -> list[str]:
    """
    Get the body of a function in an APKBUILD.

    :param path: full path to the APKBUILD
    :param func: name of function to get the body of.
    :returns: function body in an array of strings.
    """
    func_body = []
    in_func = False
    lines = read_file(path)
    for line in lines:
        if in_func:
            if line.startswith("}"):
                in_func = False
                break
            func_body.append(line)
            continue
        else:
            if line.startswith(func + "() {"):
                in_func = True
                continue
    return func_body


def read_file(path: Path) -> list[str]:
    """
    Read an APKBUILD file

    :param path: full path to the APKBUILD
    :returns: contents of an APKBUILD as a list of strings
    """
    with path.open(encoding="utf-8") as handle:
        lines = handle.readlines()
        if handle.newlines != "\n":
            raise NonBugError(f"Wrong line endings in APKBUILD: {path}")
    return lines


def parse_next_attribute(
    lines: list[str], i: int, path: Path
) -> tuple[str, str, int] | tuple[None, None, int]:
    """
    Parse one attribute from the APKBUILD.

    It may be written across multiple lines, use a quoting sign and/or have
    a comment at the end. Some examples:

    pkgrel=3
    options="!check" # ignore this comment
    arch='all !armhf'
    depends="first-pkg
    second-pkg"

    :param lines: newline-terminated list of lines from the APKBUILD
    :param i: index of the line we are currently looking at
    :param path: full path to the APKBUILD (for error message)
    :returns: (attribute, value, i)
              attribute: attribute name if any was found in line i / None
              value: that was parsed from the line
              i: line that was parsed last
    """
    # Check for and cut off "attribute="
    rematch5 = revar5.match(lines[i])
    if not rematch5:
        return (None, None, i)
    attribute = rematch5.group(0)
    value = lines[i][len(attribute) : -1]
    attribute = rematch5.group(0).rstrip("=")

    # Determine end quote sign
    end_char = None
    for char in ["'", '"']:
        if value.startswith(char):
            end_char = char
            value = value[1:]
            break

    # Single line
    if not end_char:
        value = value.split("#")[0].rstrip()
        return (attribute, value, i)
    if end_char in value:
        value = value.split(end_char, 1)[0]
        return (attribute, value, i)

    # Parse lines until reaching end quote
    i += 1
    while i < len(lines):
        line = lines[i]
        value += " "
        if end_char in line:
            value += line.split(end_char, 1)[0].strip()
            return (attribute, value.strip(), i)
        value += line.strip()
        i += 1

    raise NonBugError(
        f"Can't find closing quote sign ({end_char}) for attribute '{attribute}' in: {path}"
    )


def _parse_attributes(
    path: Path, lines: list[str], apkbuild_attributes: dict[str, dict[str, bool]], ret: Apkbuild
) -> None:
    """
    Parse attributes from a list of lines. Variables are replaced with values
    from ret (if found) and split into the format configured in
    apkbuild_attributes.

    :param lines: the lines to parse
    :param apkbuild_attributes: the attributes to parse
    :param ret: a dict to update with new parsed variable
    """
    # Parse all variables first, and replace variables mentioned earlier
    for i in range(len(lines)):
        attribute, value, i = parse_next_attribute(lines, i, path)
        if not attribute or not value:
            continue
        ret[attribute] = replace_variable(ret, value)

    if "subpackages" in apkbuild_attributes:
        subpackages: OrderedDict[str, str] = OrderedDict()
        for subpkg in ret["subpackages"].split(" "):
            if subpkg:
                _parse_subpackage(path, lines, ret, subpackages, subpkg)
        ret["subpackages"] = subpackages

    # Split attributes
    for attribute, options in apkbuild_attributes.items():
        if options.get("array", False):
            # Split up arrays, delete empty strings inside the list
            ret[attribute] = list(filter(None, ret[attribute].split()))
        if options.get("int", False):
            if ret[attribute]:
                ret[attribute] = int(ret[attribute])
            else:
                ret[attribute] = 0

    # Remove variables not in attributes
    for attribute in list(ret.keys()):
        if attribute not in apkbuild_attributes:
            del ret[attribute]


def _parse_subpackage(
    path: Path, lines: list[str], apkbuild: Apkbuild, subpackages: dict[str, Any], subpkg: str
) -> None:
    """
    Attempt to parse attributes from a subpackage function.
    This will attempt to locate the subpackage function in the APKBUILD and
    update the given attributes with values set in the subpackage function.

    :param path: path to APKBUILD
    :param lines: the lines to parse
    :param apkbuild: dict of attributes already parsed from APKBUILD
    :param subpackages: the subpackages dict to update
    :param subpkg: the subpackage to parse
                   (may contain subpackage function name separated by :)
    """
    subpkgparts = subpkg.split(":")
    subpkgname = subpkgparts[0]
    subpkgsplit = subpkgname[subpkgname.rfind("-") + 1 :]
    # If there are multiple sections to the subpackage, the middle one (item 1 in the
    # sequence in this case) is the custom function name which we should use instead of
    # the deduced one. For example:
    #
    #   "$pkgname-something-subpkg:something_subpkg:noarch"
    #
    # But only actually use it if the custom function is not an empty string, as in
    # those cases it is not meant to be set here. For example:
    #
    #   "$pkgname-something::noarch"
    #
    if len(subpkgparts) > 1 and subpkgparts[1] != "":
        subpkgsplit = subpkgparts[1]

    # Find start and end of package function
    start = end = 0
    prefix = subpkgsplit + "() {"
    for i in range(len(lines)):
        if lines[i].startswith(prefix):
            start = i + 1
        elif start and lines[i].startswith("}"):
            end = i
            break

    if not start:
        # Unable to find subpackage function in the APKBUILD.
        # The subpackage function could be actually missing, or this is a
        # problem in the parser. For now we also don't handle subpackages with
        # default functions (e.g. -dev or -doc).
        # In the future we may want to specifically handle these, and throw
        # an exception here for all other missing subpackage functions.
        subpackages[subpkgname] = None
        logging.verbose(
            f"{apkbuild['pkgname']}: subpackage function '{subpkgsplit}' for "
            f"subpackage '{subpkgname}' not found, ignoring"
        )
        return

    if not end:
        raise NonBugError(
            f"Could not find end of subpackage function, no line starts with "
            f"'}}' after '{prefix}' in {path}"
        )

    lines = lines[start:end]
    # Strip tabs before lines in function
    lines = [line.strip() + "\n" for line in lines]

    # Copy variables
    apkbuild = apkbuild.copy()
    apkbuild["subpkgname"] = subpkgname
    # Don't inherit pmb_recommends from the top-level package.
    # There are two reasons for this:
    # 1) the subpackage may specify its own pmb_recommends
    # 2) the top-level package may list the subpackage as a pmb_recommends,
    #    thereby creating a circular dependency
    apkbuild["_pmb_recommends"] = ""

    # Parse relevant attributes for the subpackage
    _parse_attributes(path, lines, pmb.config.apkbuild_package_attributes, apkbuild)

    # Return only properties interesting for subpackages
    ret = {}
    for key in pmb.config.apkbuild_package_attributes:
        ret[key] = apkbuild[key]
    subpackages[subpkgname] = ret


def _apkbuild_from_lines(
    lines: list[str], path: Path, check_pkgver: bool = True, check_pkgname: bool = True
) -> Apkbuild:
    """
    See `apkbuild()` in this module for full explanation. This function exists
    as a means of avoiding reading the file twice if you already have the
    file parsed into this format.

    :param lines: lines of the APKBUILD, as parsed by `read_file()`.
    :param path: full path to the APKBUILD, for error messages.
    :param check_pkgver: verify that the pkgver is valid.
    :param check_pkgname: the pkgname must match the name of the aport folder
    :returns: relevant variables from the APKBUILD. Arrays get returned as
              arrays.
    """
    # Parse all attributes from the config
    ret = dict.fromkeys(pmb.config.apkbuild_attributes, "")
    _parse_attributes(path, lines, pmb.config.apkbuild_attributes, ret)

    # Sanity check: pkgname
    suffix = f"/{ret['pkgname']}/APKBUILD"
    if check_pkgname and not os.path.realpath(path).endswith(suffix):
        logging.info(f"Folder: '{os.path.dirname(path)}'")
        logging.info(f"Pkgname: '{ret['pkgname']}'")
        raise NonBugError(
            "The pkgname must be equal to the name of the folder that contains the APKBUILD!"
        )

    # Sanity check: pkgver
    if check_pkgver and not pmb.parse.version.validate(ret["pkgver"]):
        logging.info(
            "NOTE: Valid pkgvers are described here: "
            "https://wiki.alpinelinux.org/wiki/APKBUILD_Reference#pkgver"
        )
        raise NonBugError(f"Invalid pkgver '{ret['pkgver']}' in APKBUILD: {path}")

    return ret


@Cache("path")
def apkbuild(path: Path, check_pkgver: bool = True, check_pkgname: bool = True) -> Apkbuild:
    """
    Parse relevant information out of the APKBUILD file. This is not meant
    to be perfect and catch every edge case (for that, a full shell parser
    would be necessary!). Instead, it should just work with the use-cases
    covered by pmbootstrap and not take too long.
    Run 'pmbootstrap apkbuild_parse hello-world' for a full output example.

    :param path: full path to the APKBUILD
    :param check_pkgver: verify that the pkgver is valid.
    :param check_pkgname: the pkgname must match the name of the aport folder
    :returns: relevant variables from the APKBUILD. Arrays get returned as
              arrays.
    """
    if path.name != "APKBUILD":
        path = path / "APKBUILD"

    if not path.exists():
        raise FileNotFoundError(f"{path.relative_to(get_context().config.work)} not found!")

    # Read the file and check line endings
    lines = read_file(path)

    return _apkbuild_from_lines(lines, path, check_pkgver, check_pkgname)


def kernels(device: str) -> dict[str, str] | None:
    """
    Get the possible kernels from a device-* APKBUILD.

    :param device: the device name, e.g. "lg-mako"
    :returns: None when the kernel is hardcoded in depends
    :returns: kernel types and their description (as read from the subpackages)
              possible types: "downstream", "stable", "mainline"
              example: {"mainline": "Mainline description", "downstream": "Downstream description"}
    """
    # Read the APKBUILD
    apkbuild_path = pmb.helpers.devices.find_path(device, "APKBUILD")
    if apkbuild_path is None:
        return None
    subpackages = apkbuild(apkbuild_path)["subpackages"]

    # Read kernels from subpackages
    ret = {}
    subpackage_prefix = f"device-{device}-kernel-"
    for subpkgname, subpkg in subpackages.items():
        if not subpkgname.startswith(subpackage_prefix):
            continue
        if subpkg is None:
            raise NonBugError(f"Cannot find subpackage function for: {subpkgname}")
        name = subpkgname[len(subpackage_prefix) :]
        ret[name] = subpkg["pkgdesc"]

    # Return
    if ret:
        return ret
    return None


def _parse_comment_tags(lines: list[str], tag: str) -> list[str]:
    """
    Parse tags defined as comments in a APKBUILD file. This can be used to
    parse e.g. the maintainers of a package (defined using # Maintainer:).

    :param lines: lines of the APKBUILD
    :param tag: the tag to parse, e.g. Maintainer
    :returns: array of values of the tag, one per line
    """
    prefix = f"# {tag}:"
    return [line[len(prefix) :].strip() for line in lines if line.startswith(prefix)]


def maintainers(path: Path) -> list[str] | None:
    """
    Parse maintainers of an APKBUILD file. They should be defined using
    # Maintainer: (first maintainer) and # Co-Maintainer: (additional
    maintainers).

    :param path: full path to the APKBUILD
    :returns: array of (at least one) maintainer, or None
    """
    lines = read_file(path)
    apkbuild = _apkbuild_from_lines(lines, path, False, False)
    maintainers = []
    if maintainer := apkbuild.get("maintainer"):
        maintainers.append(maintainer)
    maintainers += _parse_comment_tags(lines, "Maintainer")

    match len(maintainers):
        case 0:
            return None
        case 1:
            pass
        case _:
            # An APKBUILD should only have one or zero maintainer comments or fields,
            # in pmaports others should be defined using Co-Maintainer:
            raise NonBugError('Multiple Maintainer: or maintainer="" lines in APKBUILD')

    maintainers += _parse_comment_tags(lines, "Co-Maintainer")
    if "" in maintainers:
        raise NonBugError("Empty (Co-)Maintainer: tag")
    return maintainers


def archived(path: Path) -> str | None:
    """
    Return if (and why) an APKBUILD might be archived. This should be
    defined using a # Archived: <reason> tag in the APKBUILD.

    :param path: full path to the APKBUILD
    :returns: reason why APKBUILD is archived, or None
    """
    archived = _parse_comment_tags(read_file(path), "Archived")
    if not archived:
        return None
    return "\n".join(archived)



# Copyright 2026 Stefan Hansson, Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import pmb.chroot


def zap(
    dry: bool,
    http: bool,
    distfiles: bool,
    pkgs_local: bool,
    pkgs_local_mismatch: bool,
    pkgs_online_mismatch: bool,
    rust: bool,
    netboot: bool,
) -> None:
    pmb.chroot.zap(
        dry=dry,
        http=http,
        distfiles=distfiles,
        pkgs_local=pkgs_local,
        pkgs_local_mismatch=pkgs_local_mismatch,
        pkgs_online_mismatch=pkgs_online_mismatch,
        rust=rust,
        netboot=netboot,
    )

    # Don't write the "Done" message
    pmb.helpers.logging.disable()
